import asyncio
import time
from typing import Union
from dataclasses import dataclass, field
import pandas as pd

from nirvana.ops.base import BaseOperation, BaseOpOutputs
from nirvana.ops.scan import ScanOperation, ScanOpOutputs
from nirvana.ops.map import MapOperation, MapOpOutputs
from nirvana.ops.filter import FilterOperation, FilterOpOutputs
from nirvana.ops.rank import RankOperation, RankOpOutputs
from nirvana.ops.reduce import ReduceOperation, ReduceOpOutputs
from nirvana.ops.join import JoinOperation, JoinOpOutputs

OpOutputsType = Union[
    BaseOpOutputs,
    ScanOpOutputs,
    MapOpOutputs, 
    FilterOpOutputs, 
    ReduceOpOutputs, 
    JoinOpOutputs,
    RankOpOutputs
]

OpMapping = {
    "scan": ScanOperation,
    "map": MapOperation,
    "filter": FilterOperation,
    "rank": RankOperation,
    "reduce": ReduceOperation,
    "join": JoinOperation,
}


schema_mapping = {
    "scan": "SCAN({source}):[]->[{output_columns}]",
    "map": "MAP({user_instruction}):[{input_columns}]->[{output_columns}]",
    "filter": "FILTER({user_instruction}):[{input_columns}]->[Bool]",
    "rank": "RANK({user_instruction}):[{input_column}]->[Rank]",
    "reduce": "AGGR({user_instruction}):[{input_column}]->[Aggr]",
    "join": "{how}-JOIN({user_instruction}):[left[{left_on}] * right[{right_on}]",
}


@dataclass
class NodeFields:
    left_input_fields: list[str] = field(default_factory=list)
    right_input_fields: list[str] = field(default_factory=list)
    output_fields: list[str] = field(default_factory=list)


@dataclass
class DataSources:
    left_datasource: str = field(default="")
    right_datasource: str = field(default="")
    output_datasource: str = field(default="")


@dataclass
class NodeOutput:
    output: pd.DataFrame = field(default=None)
    cost: float = field(default=0.0)


class NodeBase:
    def __init__(self):
        self._left_child: NodeBase | None = None
        self._right_child: NodeBase | None = None

    @property
    def left_child(self):
        return self._left_child
    
    @property
    def right_child(self):
        return self._right_child
    
    def set_left_child(self, node: "NodeBase"):
        self._left_child = node

    def set_right_child(self, node: "NodeBase"):
        self._right_child = node


class LineageNode(NodeBase):
    def __init__(
            self, 
            op_name: str, 
            op_kwargs: dict,
            node_fields: dict,
            datasource: pd.DataFrame | None = None,
            **kwargs
    ):
        super().__init__()
        self.operator: BaseOperation = OpMapping[op_name](**op_kwargs)
        self.node_fields = NodeFields(**node_fields)
        self.datasource = datasource

    @property
    def op_name(self) -> str:
        return self.operator.op_name

    async def execute_operation(self, input: pd.DataFrame | list[pd.DataFrame] | None = None) -> OpOutputsType:
        if self.op_name == "scan":
            return await self.operator.execute(input_data=self.datasource)
        elif self.op_name == "join":
            return await self.operator.execute(left_data=input[0], right_data=input[1])
        else:
            return await self.operator.execute(input_data=input)

    async def collate_dataframe(self, input: pd.DataFrame | list[pd.DataFrame] | None = None, op_outputs: OpOutputsType | None = None) -> pd.DataFrame:
        if self.op_name == "scan":
            return self.datasource
        elif self.op_name == "join":
            input[0]["keys"] = op_outputs.left_join_keys
            input[1]["keys"] = op_outputs.right_join_keys
            output = input[0].join(input[1], on="keys", how=self.operator.how).drop("keys", axis=1)
            return output
        elif self.op_name == "filter":
            if op_outputs.output is None:
                return input
            return input[op_outputs.output]
        elif self.op_name == "map":
            if op_outputs.outputs:
                output = pd.concat([input, pd.DataFrame(op_outputs.outputs)], axis=1)
                return output
            else:
                return input
        elif self.op_name == "rank":
            return input.iloc[op_outputs.ranked_indices]
        elif self.op_name == "reduce":
            return pd.DataFrame({"reduce_result": [op_outputs.output]})

    async def run(self, input: pd.DataFrame | list[pd.DataFrame] | None = None) -> NodeOutput:
        if self.op_name == "scan":
            op_outputs = await self.operator.execute(input_data=self.datasource)
            return NodeOutput(output=op_outputs.output, cost=op_outputs.cost)
        
        elif self.op_name == "join":
            op_outputs = await self.operator.execute(left_data=input[0], right_data=input[1])
            input[0]["keys"] = op_outputs.left_join_keys
            input[1]["keys"] = op_outputs.right_join_keys
            output = input[0].join(input[1], on="keys", how=self.operator.how).drop("keys", axis=1)
            return NodeOutput(output=output, cost=op_outputs.cost)

        elif self.op_name == "filter":
            op_outputs = await self.operator.execute(input_data=input)
            if op_outputs.output is None:
                return NodeOutput(output=input, cost=op_outputs.cost)
            return NodeOutput(output=input[op_outputs.output], cost=op_outputs.cost)
        
        elif self.op_name == "map":
            op_outputs = await self.operator.execute(input_data=input)
            if op_outputs.outputs:
                output = pd.concat([input, pd.DataFrame(op_outputs.outputs)], axis=1)
                return NodeOutput(output=output, cost=op_outputs.cost)
            else:
                return NodeOutput(output=input, cost=op_outputs.cost)
        
        elif self.op_name == "rank":
            op_outputs = await self.operator.execute(input_data=input)
            output = input.iloc[op_outputs.ranked_indices]
            return NodeOutput(output=output, cost=op_outputs.cost)
        
        elif self.op_name == "reduce":
            op_outputs = await self.operator.execute(input_data=input)
            return NodeOutput(output=pd.DataFrame({"reduce_result": [op_outputs.output]}), cost=op_outputs.cost)
        
        else:
            raise ValueError(f"Unsupported operator: {self.op_name}")


def collect_op_metadata(op_node: LineageNode, max_instruction_print_length: int = 256):
    op_name = op_node.op_name
    op_kwargs = op_node.operator.op_kwargs
    user_instruction = op_kwargs["user_instruction"][:max_instruction_print_length]
    if op_name == "map":
        dependencies = op_kwargs["input_columns"][:5]
        dependencies_str = ", ".join(dependencies) + ", ..." if len(dependencies) > 5 else ", ".join(dependencies)
        generated_fields = op_kwargs["output_columns"][:5]
        generated_fields_str = ", ".join(generated_fields) + ", ..." if len(generated_fields) > 5 else ", ".join(generated_fields)
        return (
            schema_mapping[op_name].format(user_instruction=user_instruction, input_columns=dependencies_str, output_columns=generated_fields_str)
        )
    elif op_name == "filter":
        dependencies = op_kwargs["input_columns"][:5]
        dependencies_str = ", ".join(dependencies) + ", ..." if len(dependencies) > 5 else ", ".join(dependencies)
        return (
            schema_mapping[op_name].format(user_instruction=user_instruction, input_columns=dependencies_str)
        )
    elif op_name in ["rank", "reduce"]:
        return (
            schema_mapping[op_name].format(user_instruction=user_instruction, input_column=op_kwargs['input_columns'][0])
        )
    elif op_name == "join":
        return (
            schema_mapping[op_name].format(how=op_kwargs['how'], user_instruction=user_instruction, left_on=op_kwargs['left_on'][0], right_on=op_kwargs['right_on'][0])
        )
    elif op_name == "scan":
        output_columns = op_kwargs["output_columns"][:2]
        output_columns_str = ", ".join(output_columns) + ", ..."
        return (
            schema_mapping[op_name].format(source=op_kwargs['source'], output_columns=output_columns_str)
        )
    return ""


async def execute_node(node: LineageNode) -> tuple[pd.DataFrame, float]:
    if node.left_child:
        left_node_output, cost_from_left_subtree = await execute_node(node.left_child)
    if node.right_child:
        right_node_output, cost_from_right_subtree = await execute_node(node.right_child)
    
    if node.op_name == "scan":
        node_output = await node.run()
        accumulated_cost = node_output.cost
        return node_output.output, accumulated_cost
    elif node.op_name == "join":
        node_output = await node.run([left_node_output, right_node_output])
        accumulated_cost = cost_from_left_subtree + cost_from_right_subtree + node_output.cost  
        return node_output.output, accumulated_cost
    else:
        node_output = await node.run(left_node_output)
        accumulated_cost = cost_from_left_subtree + node_output.cost
        return node_output.output, accumulated_cost



import warnings

import functools
import asyncio
from typing import Any, Iterable, Callable, Literal
from dataclasses import dataclass
import pandas as pd

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.executors.tools import BaseTool, FunctionCallTool
from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.filter_prompter import FilterPrompter


def filter_wrapper(
    input_data: pd.DataFrame, 
    user_instruction: str = None,
    input_columns: list[str] = None,
    func: Callable = None,
    context: list[dict] | str | None = None,
    model: str | None = None,
    strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
    limit: int | None = None,
    rate_limit: int = 16,
    assertions: list[Callable] | None = [],
    **kwargs
):
    """
    A function wrapper for filter operation

    Args:
        input_data (pd.DataFrame): Input dataframe
        user_instruction (str, optional): User instruction. Defaults to None.
        input_columns (list[str], optional): Input columns. Defaults to None.
        func (Callable, optional): User function. Defaults to None.
        context (list[dict] | str, optional): Context. Defaults to None.
        model (str, optional): Model. Defaults to None.
        strategy (Literal["plain", "fewshot", "self-refine"], optional): Strategy. Defaults to "plain".
        limit (int): Maximum number of outputs to produce before stopping.
        rate_limit (int, optional): Rate limit. Defaults to 16.
        assertions (list[Callable], optional): Assertions. Defaults to [].
        **kwargs: Additional keyword arguments for OpenAI Clent.
    """
    
    filter_op = FilterOperation(
        user_instruction=user_instruction,
        input_columns=input_columns,
        context=context,
        model=model,
        tool=FunctionCallTool.from_function(func=func) if func else None,
        strategy=strategy,
        limit=limit,
        rate_limit=rate_limit,
        assertions=assertions,
    )
    outputs = asyncio.run(filter_op.execute(
        input_data=input_data,
        **kwargs
    ))
    return outputs


@dataclass
class FilterOpOutputs(BaseOpOutputs):
    output: Iterable[bool] = None

    def __add__(self, other: "FilterOpOutputs"):
        return FilterOpOutputs(
            output=self.output + other.output,
            cost=self.cost + other.cost
        )


class FilterOperation(BaseOperation):
    """
    Filter operator: Uses an LLM to evaluate a natural language predicate on a column
    """
    strategy_options = ["plain", "fewshot", "self_refine"]
    
    def __init__(
        self,
        user_instruction: str = "",
        input_columns: list[str] = [],
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: Callable | BaseTool | None = None,
        strategy: Literal["plain", "fewshot", "self-refine"] = "plain",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        if tool and not isinstance(tool, BaseTool):
            tool = FunctionCallTool.from_function(func=tool)
        
        super().__init__(
            op_name="filter",
            user_instruction=user_instruction,
            context=context,
            model=model,
            tool=tool,
            strategy=strategy,
            limit=limit,
            rate_limit=rate_limit,
            assertions=assertions,
        )
        self.prompter = FilterPrompter()
        self.input_columns = input_columns

    @property
    def dependencies(self) -> list[str]:
        return self.input_columns
    
    @property
    def generated_fields(self) -> list[str]:
        return []
    
    @property
    def op_kwargs(self) -> dict:
        kwargs = super().op_kwargs
        kwargs["input_columns"] = self.input_columns
        return kwargs

    async def _execute_by_plain_llm(self, data: pd.Series, user_instruction: str, dtypes: list[str], **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_prompt(data, user_instruction, dtypes)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            result = self._postprocess_filter_output(output["output"])
            return result, output["cost"]

    async def _execute_by_fewshot_llm(self, data: pd.Series, user_instruction: str, dtypes: list[str], demos, **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_fewshot_prompt(data, user_instruction, dtypes, demos)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            result = self._postprocess_filter_output(output["output"])
            return result, output["cost"]

    async def _execute_by_self_refine(self, data: pd.Series, user_instruction: str, dtypes: list[str], **kwargs):
        async with self.semaphore:
            self_refine_cost = 0.0
            generate_prompt = self.prompter.generate_prompt(data, user_instruction, dtypes)
            output = await self.llm(generate_prompt, parse_tags=True, tags=["output"], **kwargs)
            self_refine_cost += output["cost"]

            evaluate_prompt = self.prompter.generate_evaluate_prompt(data, output["raw_output"], user_instruction, dtypes)
            evaluate_output = await self.llm(evaluate_prompt, parse_tags=True, tags=["evaluation", "feedback"], **kwargs)
            self_refine_cost += evaluate_output["cost"]
            
            if "pass" in evaluate_output["evaluation"].lower():
                result = self._postprocess_filter_output(output["output"])
                return result, self_refine_cost
            else:
                refine_prompt = self.prompter.generate_refine_prompt(data, output["raw_output"], user_instruction, evaluate_output["feedback"], dtypes)
                refine_output = await self.llm(refine_prompt, parse_tags=True, tags=["output"], **kwargs)
                self_refine_cost += refine_output["cost"]
                result = self._postprocess_filter_output(refine_output["output"])
                return result, self_refine_cost

    async def _execute_by_func(self, data: pd.Series, user_instruction: str, func: Callable, llm_call: Callable, **kwargs):
        try:
            output = func(data)
            return output, 0.0
        except Exception as e:
            return await llm_call(data, user_instruction)
        
    def _postprocess_filter_output(self, llm_result: str | bool | None) -> bool:
        if llm_result is None:
            return False
        if isinstance(llm_result, bool):
            return llm_result
        if "true" in llm_result.lower():
            return True
        elif "false" in llm_result.lower():
            return False
        else:
            return False

    async def execute(
        self, 
        input_data: pd.DataFrame,
        **kwargs
    ):
        if self.user_instruction is None and not self.has_udf():
            raise ValueError("Neither `user_instruction` nor `func` is given.")
        
        if input_data.empty:
            return FilterOpOutputs()
        
        processed_data = input_data[self.input_columns]
        dtypes = []
        for col in self.input_columns:
            if isinstance(input_data[col].dtype, ImageDtype):
                dtypes.append("image")
            else:
                dtypes.append("text")

        if self.strategy == "plain":
            execution_func = functools.partial(self._execute_by_plain_llm, dtypes=dtypes, field_name=self.input_columns[0], model=self.model, **kwargs)
        elif self.strategy == "fewshot":
            assert self.context is not None, "Few-shot examples must be provided in the context for in-context learning."
            demos = self.context
            execution_func = functools.partial(self._execute_by_fewshot_llm, dtypes=dtypes, demos=demos, field_name=self.input_columns[0], model=self.model, **kwargs)
        elif self.strategy == "self_refine":
            execution_func = functools.partial(self._execute_by_self_refine, dtypes=dtypes, field_name=self.input_columns[0], model=self.model, **kwargs)
        else:
            raise ValueError(f"The optional strategies available for filter are {self.strategy_options}. Strategy {self.strategy} is not supported.")

        # Create tasks for all data points
        tasks = []
        for _, data in processed_data.iterrows():
            if data.empty:
                tasks.append(asyncio.create_task(asyncio.sleep(0, result=(False, 0.0))))
            elif self.has_udf():
                tasks.append(asyncio.create_task(self._execute_by_func(data, self.user_instruction, self.tool, execution_func)))
            else:
                tasks.append(asyncio.create_task(execution_func(data, self.user_instruction)))

        # Wait for all tasks to complete
        if self.limit is not None and self.limit <= 0:
            warnings.warn("The limit should be positive. To execute, the limit will be ignored.")
            self.limit = None
        
        token_cost = 0.0
        filter_outputs: list[bool] = []
        if self.limit is not None:
            num_passed_records: int = 0
            reach_limit: bool = False
            for i in range(0, len(tasks), self.limit):
                if reach_limit:
                    break
                batch_tasks = tasks[i:i + self.limit]
                batch_results = await asyncio.gather(*batch_tasks)
                token_cost += sum([result[1] for result in batch_results])
                for result, _ in batch_results:
                    filter_outputs.append(result)
                    if result:
                        num_passed_records += 1
                    if num_passed_records >= self.limit:
                        reach_limit = True
                        break
            num_remaining_records = len(processed_data) - len(filter_outputs)
            if num_remaining_records > 0:
                filter_outputs.extend([False] * num_remaining_records)
        else:
            results = await asyncio.gather(*tasks)
            filter_outputs = [result[0] for result in results]
            token_cost = sum([result[1] for result in results])
        
        return FilterOpOutputs(
            output=filter_outputs,
            cost=token_cost
        )



import pytest
import pandas as pd
from unittest.mock import AsyncMock, MagicMock, patch
import asyncio

import nirvana as nv
from nirvana.ops.map import MapOperation, MapOpOutputs
from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import FunctionCallTool


@pytest.fixture
def sample_dataframe():
    """Create a sample dataframe for testing."""
    return pd.DataFrame({
        "title": ["The Godfather", "The Dark Knight", "Inception"],
        "overview": [
            "An organized crime dynasty's aging patriarch transfers control of his clandestine empire to his reluctant son.",
            "When the menace known as the Joker wreaks havoc and chaos on the people of Gotham, Batman must accept one of the greatest psychological and physical tests of his ability to fight injustice.",
            "A thief who steals corporate secrets through the use of dream-sharing technology is given the inverse task of planting an idea into the mind of a C.E.O."
        ]
    })


@pytest.fixture
def empty_dataframe():
    """Create an empty dataframe for testing."""
    return pd.DataFrame({"title": [], "overview": []})


@pytest.fixture
def dataframe_with_nan():
    """Create a dataframe with NaN values for testing."""
    return pd.DataFrame({
        "title": ["The Godfather", None, "Inception"],
        "overview": [
            "An organized crime dynasty's aging patriarch transfers control of his clandestine empire to his reluctant son.",
            None,
            "A thief who steals corporate secrets through the use of dream-sharing technology is given the inverse task of planting an idea into the mind of a C.E.O."
        ]
    })


@pytest.fixture
def mock_llm_client():
    """Create a mock LLM client for testing."""
    
    async def mock_call(messages, parse_tags=False, parse_code=False, **kwargs):
        """Mock LLM call that returns structured output."""
        result = {
            "raw_output": "<genre>crime, drama</genre>",
            "cost": 0.01
        }
        
        if parse_tags:
            tags = kwargs.get("tags", [])
            # When parse_tags is True, the LLM client extracts XML tags
            # and puts them in the result dict with tag names as keys
            for tag in tags:
                if tag == "genre":
                    result[tag] = "crime, drama"
                elif tag == "evaluation":
                    result[tag] = "PASS"
                elif tag == "feedback":
                    result[tag] = "No feedback needed"
        
        return result
    
    # Create a proper async mock
    mock_client = AsyncMock(spec=LLMClient, side_effect=mock_call)
    mock_client.default_model = "gpt-4"
    return mock_client


class TestMapOperation:
    """Test suite for MapOperation class."""
    
    def test_map_operation_initialization(self, mock_llm_client):
        """Test that MapOperation initializes correctly."""
        MapOperation.set_llm(mock_llm_client)
        op = MapOperation(
            user_instruction="Extract genre from the movie overview",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="plain"
        )
        
        assert op.op_name == "map"
        assert op.dependencies == ["overview"]
        assert op.generated_fields == ["genre"]
    
    def test_op_kwargs_property(self, mock_llm_client):
        """Test that op_kwargs includes input and output columns."""
        MapOperation.set_llm(mock_llm_client)
        op = MapOperation(
            user_instruction="Extract genre from the movie overview",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="plain"
        )
        
        kwargs = op.op_kwargs
        assert "input_columns" in kwargs
        assert "output_columns" in kwargs
        assert kwargs["input_columns"] == ["overview"]
        assert kwargs["output_columns"] == ["genre"]
    
    @pytest.mark.asyncio
    async def test_execute_with_empty_dataframe(self, mock_llm_client):
        """Test that execute returns None output for empty dataframe."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract genre from the movie overview",
            input_columns=["overview"],
            output_columns=["genre"]
        )
        
        empty_df = pd.DataFrame({"overview": []})
        result = await op.execute(input_data=empty_df)
        
        assert result.outputs == {}
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_with_plain_strategy(self, sample_dataframe, mock_llm_client):
        """Test execute with plain LLM strategy."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract the genre of each movie from the movie overview.",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="plain"
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
        assert result.cost > 0
    
    @pytest.mark.asyncio
    async def test_execute_with_self_refine_strategy(self, sample_dataframe, mock_llm_client):
        """Test execute with self-refine strategy."""
        # Mock the evaluation response
        async def mock_call_with_eval(messages, parse_tags=False, **kwargs):
            result = {
                "raw_output": "<genre>crime, drama</genre>",
                "cost": 0.01
            }
            
            if parse_tags:
                tags = kwargs.get("tags", [])
                for tag in tags:
                    if tag == "evaluation":
                        result[tag] = "PASS"
                    elif tag == "feedback":
                        result[tag] = "No feedback needed"
                    elif tag == "genre":
                        result[tag] = "crime, drama"
            
            return result
        
        mock_llm_client.__call__ = mock_call_with_eval
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract the genre of each movie.",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="self_refine"
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
    
    @pytest.mark.asyncio
    async def test_execute_with_fewshot_strategy(self, sample_dataframe, mock_llm_client):
        """Test execute with fewshot strategy."""
        MapOperation.set_llm(mock_llm_client)
        
        demos = [
            {"data": pd.Series({"overview": "A crime movie"}), "answer": "<genre>crime</genre>"},
            {"data": pd.Series({"overview": "A comedy film"}), "answer": "<genre>comedy</genre>"}
        ]
        
        op = MapOperation(
            user_instruction="Extract the genre of each movie.",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="fewshot",
            context=demos
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
    
    @pytest.mark.asyncio
    async def test_execute_with_udf(self, sample_dataframe, mock_llm_client):
        """Test execute with user-defined function."""
        MapOperation.set_llm(mock_llm_client)
        
        def extract_genre(data: pd.Series):
            """Simple function to extract genre."""
            if "crime" in data["overview"].lower():
                return {"genre": "crime"}
            elif "action" in data["overview"].lower():
                return {"genre": "action"}
            return {"genre": "unknown"}
        
        tool = FunctionCallTool.from_function(func=extract_genre)
        
        op = MapOperation(
            user_instruction="Extract the genre of each movie.",
            input_columns=["overview"],
            output_columns=["genre"],
            tool=tool
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
        # UDF should return the function results directly
        assert "crime" in result.outputs["genre"]
    
    @pytest.mark.asyncio
    async def test_execute_with_udf_exception_fallback(self, sample_dataframe, mock_llm_client):
        """Test that UDF exceptions fall back to LLM."""
        MapOperation.set_llm(mock_llm_client)
        
        def failing_func(text):
            """Function that raises an exception."""
            raise ValueError("Function failed")
        
        tool = FunctionCallTool.from_function(func=failing_func)
        
        op = MapOperation(
            user_instruction="Extract the genre of each movie.",
            input_columns=["overview"],
            output_columns=["genre"],
            tool=tool
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        # Should fall back to LLM, so result should still be valid
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
    
    @pytest.mark.asyncio
    async def test_execute_missing_user_instruction_and_func(self, sample_dataframe, mock_llm_client):
        """Test that execute raises ValueError when both user_instruction and func are missing."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction=None,
            input_columns=["overview"],
            output_columns=["genre"]
        )
        
        with pytest.raises(ValueError, match="Neither `user_instruction` nor `func` is given."):
            await op.execute(input_data=sample_dataframe)
    
    @pytest.mark.asyncio
    async def test_execute_unsupported_strategy(self, sample_dataframe, mock_llm_client):
        """Test that execute raises ValueError for unsupported strategy."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract genre",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="unsupported_strategy"
        )
        
        with pytest.raises(ValueError, match="Strategy unsupported_strategy is not supported."):
            await op.execute(input_data=sample_dataframe)
    
    @pytest.mark.asyncio
    async def test_execute_fewshot_without_context(self, sample_dataframe, mock_llm_client):
        """Test that execute raises AssertionError for fewshot without context."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract genre",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="fewshot"
        )
        
        with pytest.raises(AssertionError, match="Few-shot examples must be provided"):
            await op.execute(input_data=sample_dataframe)

    @pytest.mark.asyncio
    async def test_map_operation_with_limit(self, sample_dataframe, mock_llm_client):
        """Test that map operation respects the limit on processed rows and pads the rest with None."""
        MapOperation.set_llm(mock_llm_client)

        op = MapOperation(
            user_instruction="Extract genre from overview",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="plain",
            limit=2,
        )

        result = await op.execute(input_data=sample_dataframe)

        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        # Always one output per input row
        assert len(result.outputs["genre"]) == len(sample_dataframe)
        # Only first 2 rows should be filled, the rest padded with None
        assert result.outputs["genre"][:2] == ["crime, drama", "crime, drama"]
        assert all(v is None for v in result.outputs["genre"][2:])


class TestMapOpOutputs:
    """Test suite for MapOpOutputs class."""
    
    def test_map_op_outputs_initialization(self):
        """Test that MapOpOutputs initializes correctly with list field_name."""
        outputs = MapOpOutputs(
            outputs={"genre": ["crime", "drama"], "rating": [8.5, 9.0]},
            cost=0.05
        )
        
        assert outputs.outputs == {"genre": ["crime", "drama"], "rating": [8.5, 9.0]}
        assert outputs.cost == 0.05
    
    def test_map_op_outputs_default_values(self):
        """Test that MapOpOutputs has correct default values."""
        outputs = MapOpOutputs()
        
        assert outputs.outputs == {}
        assert outputs.cost == 0.0
    
    def test_map_op_outputs_addition_field_name(self):
        """Test that MapOpOutputs can be added together with list field_name."""
        outputs1 = MapOpOutputs(
            outputs={"genre": ["crime"], "rating": [8.5]},
            cost=0.02
        )
        
        outputs2 = MapOpOutputs(
            outputs={"genre": ["drama"], "rating": [9.0]},
            cost=0.03
        )
        
        combined = outputs1 + outputs2
        
        assert combined.outputs == {"genre": ["crime", "drama"], "rating": [8.5, 9.0]}
        assert combined.cost == 0.05
    
    def test_map_op_outputs_addition_different_field_names(self):
        """Test that adding MapOpOutputs with different field names raises AssertionError."""
        outputs1 = MapOpOutputs(outputs={"genre": ["crime"]}, cost=0.01)
        outputs2 = MapOpOutputs(outputs={"rating": ["8.5"]}, cost=0.01)
        
        with pytest.raises(AssertionError, match="Cannot merge MapOpOutputs with different field names"):
            _ = outputs1 + outputs2


class TestMapWrapper:
    """Test suite for map_wrapper function."""
    
    def test_map_wrapper_basic(self, sample_dataframe, mock_llm_client):
        """Test the map_wrapper function with basic usage."""
        MapOperation.set_llm(mock_llm_client)
        
        result = nv.ops.map(
            sample_dataframe,
            "Extract the genre of each movie.",
            input_columns=["overview"],
            output_columns=["genre"],
            strategy="plain"
        )
        
        assert isinstance(result, MapOpOutputs)
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3
    
    def test_map_wrapper_with_udf(self, sample_dataframe, mock_llm_client):
        """Test the map_wrapper function with user-defined function."""
        MapOperation.set_llm(mock_llm_client)
        
        def extract_genre(data: pd.Series):
            if "crime" in data["overview"].lower():
                return {"genre": "crime"}
            return {"genre": "other"}

        result = nv.ops.map(
            sample_dataframe,
            input_columns=["overview"],
            output_columns=["genre"],
            func=extract_genre
        )
        
        assert isinstance(result, MapOpOutputs)
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert len(result.outputs["genre"]) == 3


class TestMapOperationIntegration:
    """Integration tests for map operation with real-like scenarios."""
    
    @pytest.mark.asyncio
    async def test_map_operation_with_string_dtype(self, mock_llm_client):
        """Test map operation with string data type."""
        MapOperation.set_llm(mock_llm_client)
        
        df = pd.DataFrame({
            "text": ["Hello world", "Python programming", "Data science"]
        })
        
        op = MapOperation(
            user_instruction="Extract the main topic.",
            input_columns=["text"],
            output_columns=["topic"],
            strategy="plain"
        )
        
        result = await op.execute(input_data=df)
        
        assert isinstance(result.outputs, dict)
        assert "topic" in result.outputs
        assert len(result.outputs["topic"]) == 3
    
    @pytest.mark.asyncio
    async def test_map_operation_postprocess_outputs(self, mock_llm_client):
        """Test that _postprocess_map_output handles None values correctly."""
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract genre",
            input_columns=["overview"],
            output_columns=["genre"]
        )
        
        # Non-None result
        single = op._postprocess_map_output({"genre": "crime"}, ["genre"])
        assert single == {"genre": "crime"}

        # None result should be mapped to a dict with None values
        none_result = op._postprocess_map_output(None, ["genre"])
        assert none_result == {"genre": None}
    
    @pytest.mark.asyncio
    async def test_map_operation_multiple_output_columns(self, sample_dataframe, mock_llm_client):
        """Test map operation with multiple output columns."""
        # Mock LLM to return multiple fields
        async def mock_call_multi(messages, parse_tags=False, **kwargs):
            result = {
                "raw_output": "<genre>crime</genre><rating>9.2</rating>",
                "cost": 0.01
            }
            if parse_tags:
                tags = kwargs.get("tags", [])
                for tag in tags:
                    if tag == "genre":
                        result[tag] = "crime"
                    elif tag == "rating":
                        result[tag] = "9.2"
            return result
        
        mock_llm_client.__call__ = mock_call_multi
        MapOperation.set_llm(mock_llm_client)
        
        op = MapOperation(
            user_instruction="Extract genre and rating.",
            input_columns=["overview"],
            output_columns=["genre", "rating"],
            strategy="plain"
        )
        
        result = await op.execute(input_data=sample_dataframe)
        
        assert isinstance(result.outputs, dict)
        assert "genre" in result.outputs
        assert "rating" in result.outputs
        assert len(result.outputs["genre"]) == 3
        assert len(result.outputs["rating"]) == 3



from dataclasses import asdict
from nirvana.lineage.abstractions import LineageNode


class NonLLMPushdown:
    @classmethod
    def check_pattern(cls, node: LineageNode, dependencies: list[str], generated_fields: list[str]) -> bool:
        return node.operator.has_udf() and all([field not in generated_fields for field in dependencies])

    @classmethod
    async def push_down(cls, node: LineageNode) -> LineageNode:
        if node.op_name in ["map", "filter"]:
            last_node = await cls.push_down(node.left_child)

            if last_node.op_name in ["scan", "join"]:
                node.set_left_child(last_node)
                return node
            
            dependencies = node.operator.dependencies
            generated_fields = last_node.operator.generated_fields
            if cls.check_pattern(node, dependencies, generated_fields):
                # push non-LLM ops down if they have a UDF and their action scope is not included in the output_fields of their ancestors
                new_node = LineageNode(op_name=node.op_name, op_kwargs=node.operator.op_kwargs, node_fields=asdict(node.node_fields))
                # swap info (eg fields) of current op and its predecessor
                new_node.node_fields.left_input_fields = last_node.node_fields.left_input_fields
                new_node.node_fields.output_fields = list(
                    set(new_node.node_fields.left_input_fields + new_node.operator.generated_fields)
                )
                last_node.node_fields.left_input_fields = new_node.node_fields.output_fields
                last_node.node_fields.output_fields = list(
                    set(last_node.node_fields.left_input_fields + last_node.operator.generated_fields)
                )
                new_node.set_left_child(last_node.left_child)
                last_node.set_left_child(await cls.push_down(new_node))
                del node
                return last_node
            else:
                node.set_left_child(last_node)
                return node
        
        elif node.op_name == "join":
            node.set_left_child(await cls.push_down(node.left_child))
            node.set_right_child(await cls.push_down(node.right_child))
            return node

        elif node.op_name == "reduce":
            node.set_left_child(await cls.push_down(node.left_child))
            return node
        
        else:
            return node
    
    @classmethod
    async def transform(cls, node: LineageNode) -> tuple[LineageNode, float]:
        node = await cls.push_down(node)
        return node, 0.0



import os
import asyncio
import pandas as pd

from nirvana.datalake.lake_manager import LakeMetadata, LakeManager
from nirvana.ops.discover import SummarizeOperation, VectorizeOperation, DiscoverOperation


class DataLake:
    """
    An interface over a data lake that serves data discovery (now) and more operations (in the future) over a data lake.
    """

    def __init__(self, embedding_model):
        self.mgr = LakeManager()
        self.embedding_model = embedding_model

    def register(self, lake_dir, storage: str = "local", **kwargs):
        self.lake_dir = lake_dir
        if storage == "local":
            for path in os.listdir(lake_dir):
                if os.path.isdir(os.path.join(lake_dir, path)):
                    self.register(os.path.join(lake_dir, path), storage, **kwargs)
                elif os.path.isfile(os.path.join(lake_dir, path)):
                    self.register_data(os.path.join(lake_dir, path))
                else:
                    ValueError(f"There is no such file or directory: {os.path.join(lake_dir, path)}")
        elif storage == "s3":
            raise NotImplementedError("S3 storage will be implemented in the near future.")
        else:
            raise ValueError(f"Unsupported storage: {storage}")
        
    def _read_context(self, context_path: str):
        file_ext = os.path.splitext(context_path)[-1][1:]
        if file_ext == "txt":
            with open(context_path, "r") as f:
                return f.read()
        else:
            raise ValueError(f"Unsupported file type: {file_ext}")            

    def register_data(self, data_path, context_path: str = None):
        if context_path is not None:
            context = self._read_context(context_path)
        else:
            context = None
        metadata = LakeMetadata(data_path, context)
        self.mgr.append(metadata)

    def summarize_data(
            self,
            column_level: bool = True, 
            record_level: bool = True, 
            sample_size: int = 5,
            strategy: str = None,
            max_length: int = 4096,
            **kwargs
    ):
        summarize_op = SummarizeOperation(**kwargs)
        total_cost = 0.0
        for metadata in self.mgr:
            data_to_summarize = pd.read_csv(metadata.data_path)
            summarize_output = asyncio.run(
                summarize_op.execute(
                    data_to_summarize, 
                    column_level=column_level, 
                    record_level=record_level,
                    sample_size=sample_size,
                    strategy=strategy,
                    max_length=max_length,
                )
            )
            metadata.column_summary, metadata.record_summary = summarize_output.column_summary, summarize_output.record_summary
            metadata.is_summarized = True
            total_cost += summarize_output.cost
        return total_cost

    def generate_index(self, **kwargs):
        vectorize_op = VectorizeOperation(**kwargs)
        vectorize_output = asyncio.run(self.vectorize_op.execute(
            self.mgr,
            self.embedding_model,
        ))
        self.mgr.text_index.index(vectorize_output.lexical_vectors, show_progress=False)
        corpus = [{"id": idx, "text": doc} for idx, doc in zip(vectorize_output.doc_ids, vectorize_output.documents)]
        self.mgr.text_index.corpus = self.mgr.text_index.corpus + corpus
        self.mgr.text_index.save(
            os.path.join(self.mgr.text_index_path, "bm25s_index"),
            corpus=self.mgr.text_index.corpus
        )
        self.mgr.vector_index.add(
            embeddings=vectorize_output.semantic_vectors,
            documents=vectorize_output.documents,
            ids=vectorize_output.doc_ids
        )
        return vectorize_output.cost

    def discover_data(
        self, 
        query: str,
        topk: int = 1,
        pool_factor: int = 5,
        alpha: float = 0.5,
        **kwargs
    ):
        discover_op = DiscoverOperation(**kwargs)
        discover_output = asyncio.run(
            discover_op.execute(
                query,
                vector_index=self.mgr.vector_index,
                text_index=self.mgr.text_index,
                embedding_model=self.embedding_model,
                topk=topk,
                pool_factor=pool_factor,
                alpha=alpha
            )
        )
        pass



from typing import Any, Union, Iterable


class SummarizePrompter:
    def __init__(self):
        self.instruction =  (
            "A table has the following columns:\n/*\n{columns}\n*/\n"
            "Describe briefly what the {column} column represents. If not possible, simply state 'No description'."
        )

    def generate_prompt(
            self,
            column: str,
            columns: Union[str, Iterable[str]],
    ):
        # 1. Prepare message
        prompt = self.instruction.format(
            columns=" | ".join(columns) if isinstance(columns, Iterable) else columns,
            column=column
        )
        message = [{"role": "user", "content": prompt}]
        return message


class RerankPrompter:
    def __init__(self):
        self.content_instruction = (
            "Given a table with the following columns:\n*/\n{columns}\n*/\n"
            "and this question:\n*/{query}\n*/\n"
            "Is the table relevant to answer the question? The answer should strictly be either True or False.\n"
            "Output the answer concisely in the following format.\n"
            "<output> True or False </output>\n"
        )
        self.context_instruction = (
            "Given this context describing a table:\n*/\n{context}\n*/\n"
            "and this question:\n*/{query}\n*/\n"
            "Is the table relevant to answer the question? The answer should strictly be either True or False.\n"
            "Output the answer concisely in the following format.\n"
            "<output> True or False </output>\n"
        )
    
    def generate_prompt(
            self,
            query: str,
            columns_desc: str = None,
            context_desc: str = None,
    ):
        # 1. Prepare message
        if columns_desc:
            prompt = self.content_instruction.format(columns=columns_desc, query=query)
        elif context_desc:
            prompt = self.context_instruction.format(context=context_desc, query=query)
        else:
            raise ValueError("Either columns_desc or context_desc should be provided.")
        message = [{"role": "user", "content": prompt}]
        return message



import time
import asyncio
import logging
import functools
from typing import Coroutine
import numpy as np
import pandas as pd

from nirvana.executors.llm_backbone import LLMClient
from nirvana.lineage.abstractions import LineageNode, OpOutputsType

logger = logging.getLogger(__name__)


ORDERED_MODELS= [  # llm list ordered by speed and intelligence (weak but fast -> strong but slow)
    "gpt-4.1-nano-2025-04-14",
    "gpt-4o-mini-2024-07-18",
    "gpt-4.1-mini-2025-04-14",
    "gpt-4.1-2025-04-14",
]


def concate_output(output1: OpOutputsType, output2: OpOutputsType):
    return output1 + output2


def cosine_similarity(x: np.array, y: np.array):
    x_norms = np.linalg.norm(x, axis=1)
    y_norms = np.linalg.norm(y, axis=1)
    dot_products = np.sum(x * y, axis=1)
    return dot_products / (x_norms * y_norms)


class PhysicalOptimizer:
    def __init__(
        self,
        agent: LLMClient = None,
        available_models: list[str] = ORDERED_MODELS,
        num_samples: int = 10,
        improve_margin: float = 0.2,
    ):
        self.agent = agent
        self.available_models = available_models if available_models else ORDERED_MODELS
        self.num_samples = num_samples
        self.improve_margin = improve_margin

    def should_optimize(self, op_name: str, input_data: pd.DataFrame | list[pd.DataFrame], num_samples: int) -> bool:
        if op_name == "join":
            # and len(input_data[0]) * len(input_data[1]) < 2 * num_samples:
            # we don't optimize join yet
            return False
        elif op_name in ["filter", "map"] and len(input_data) < 2 * num_samples:
            return False
        elif op_name in ["reduce", "rank"]:
            return False
        else:
            return True

    def split_input_data(self, op_name: str, input_data: pd.DataFrame | list[pd.DataFrame], num_samples: int):
        if op_name == "join":
            left_length, right_length = len(input_data[0]), len(input_data[1])
            if left_length > right_length:
                return [input_data[0].iloc[:num_samples], input_data[1]], [input_data[0].iloc[num_samples:], input_data[1]]
            else:
                return [input_data[0], input_data[1].iloc[:num_samples]], [input_data[0], input_data[1].iloc[num_samples:]]
        else:
            return input_data.iloc[:num_samples], input_data.iloc[num_samples:]
        
    def _serialize_map_outputs(self, outputs: dict[str, list]) -> list[str]:
        serialized_outputs = []
        for values in zip(*outputs.values()):
            values_with_key = map(lambda k, v: f"{k}: {v}", outputs.keys(), values)
            serialized_outputs.append(", ".join(values_with_key))
        return serialized_outputs

    async def optimize_exec_model(
        self, 
        node: LineageNode, 
        input_data: pd.DataFrame | list[pd.DataFrame],
    ):
        if not self.should_optimize(node.op_name, input_data, self.num_samples):
            node.operator.model = self.available_models[0] if node.op_name in ["join", "rank"] else self.available_models[-1]
            node_output = await node.execute_operation(input_data)
            return node_output

        optimize_start_time = time.time()
        train_set, test_set = self.split_input_data(node.op_name, input_data, self.num_samples)
        exec_model = self.available_models[0]
        node.operator.model = exec_model 
        node_output = None

        if node.op_name == "filter":
            worst_output = await node.execute_operation(train_set)
            node_output = worst_output
            
            improve_score_list = [0.0]
            mismatch_idx_worst2subworst = None
            mismatch_idx_subworst2subbest = None
            for model_id in range(1, len(ORDERED_MODELS)):
                if model_id == 1:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    better_output = await node.execute_operation(train_set)
                    node_output.cost += better_output.cost
                    subworst_output = np.array(better_output.output)
                    mismatch_idx_worst2subworst = np.array(worst_output.output) != subworst_output
                    mismatch_record_ratio = sum(mismatch_idx_worst2subworst) / len(worst_output.output)
                    improve_score_list.append(mismatch_record_ratio)

                elif model_id == 2:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    better_output = await node.execute_operation(train_set)
                    node_output.cost += better_output.cost
                    subbest_output = np.array(better_output.output)
                    mismatch_idx_subworst2subbest = subworst_output != subbest_output
                    if mismatch_idx_worst2subworst.all():
                        mismatch_ratio = 0.0
                    else:
                        mismatch_ratio = (
                            (mismatch_idx_subworst2subbest & ~mismatch_idx_worst2subworst).sum() / (~mismatch_idx_worst2subworst).sum()
                        )
                    improve_score = improve_score_list[-1] + mismatch_ratio * (1 - improve_score_list[-1])
                    improve_score_list.append(improve_score)

                elif model_id == 3:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    all_matched_idx = ~mismatch_idx_subworst2subbest & ~mismatch_idx_worst2subworst
                    if all_matched_idx.any():
                        better_output = await node.execute_operation(train_set.iloc[all_matched_idx])
                        node_output.cost += better_output.cost
                        best_output = np.array(better_output.output)
                        mismatch_idx_subbest2best = subbest_output[all_matched_idx] != best_output
                        mismatch_ratio = mismatch_idx_subbest2best.sum() / all_matched_idx.sum()
                        subbest_output[all_matched_idx] = best_output
                        better_output.output = subbest_output.tolist()
                    else:
                        mismatch_ratio = 0.0
                    if mismatch_idx_worst2subworst.any():
                        matched_ratio = (
                            (~mismatch_idx_subworst2subbest & mismatch_idx_worst2subworst).sum() / mismatch_idx_worst2subworst.sum()
                        )
                    else:
                        matched_ratio = 0.0
                    improve_score = mismatch_ratio * (1 - improve_score_list[-1]) + improve_score_list[-1] - improve_score_list[-2] + matched_ratio
                    improve_score_list.append(improve_score)

                if improve_score_list[-1] - improve_score_list[-2] > self.improve_margin:
                    exec_model = better_model
                    node_output.output = better_output.output

        elif node.op_name == "map":
            worst_output = await node.execute_operation(train_set)
            node_output = worst_output
            serialized_outputs = self._serialize_map_outputs(worst_output.outputs)
            worst_output_embeds, embed_cost = await self.agent.create_embedding(serialized_outputs)
            node_output.cost += embed_cost

            improve_score_list = [0.0]
            mismatch_idx_worst2subworst = None
            mismatch_idx_subworst2subbest = None
            for model_id in range(1, len(ORDERED_MODELS)):
                if model_id == 1:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    better_output = await node.execute_operation(train_set)
                    node_output.cost += better_output.cost
                    serialized_outputs = self._serialize_map_outputs(better_output.outputs)
                    subworst_output_embeds, embed_cost = await self.agent.create_embedding(serialized_outputs)
                    node_output.cost += embed_cost
                    cos_sim = cosine_similarity(worst_output_embeds, subworst_output_embeds)
                    mismatch_idx_worst2subworst = cos_sim < 0.5
                    mismatch_ratio = sum(mismatch_idx_worst2subworst) / len(serialized_outputs)
                    improve_score_list.append(mismatch_ratio)
                
                elif model_id == 2:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    better_output = await node.execute_operation(train_set)
                    node_output.cost += better_output.cost
                    subbest_output = np.array(list(better_output.outputs.values()))
                    serialized_outputs = self._serialize_map_outputs(better_output.outputs)
                    subbest_output_embeds, embed_cost = await self.agent.create_embedding(serialized_outputs)
                    node_output.cost += embed_cost
                    cos_sim = cosine_similarity(subworst_output_embeds, subbest_output_embeds)
                    mismatch_idx_subworst2subbest = cos_sim < 0.5
                    if mismatch_idx_worst2subworst.all():
                        mismatch_ratio = 0.0
                    else:
                        mismatch_ratio = (
                            (mismatch_idx_subworst2subbest & ~mismatch_idx_worst2subworst).sum() / (~mismatch_idx_worst2subworst).sum()
                        )
                    improve_score = improve_score_list[-1] + mismatch_ratio * (1 - improve_score_list[-1])
                    improve_score_list.append(improve_score)
                
                elif model_id == 3:
                    better_model = ORDERED_MODELS[model_id]
                    node.operator.model = better_model
                    all_matched_idx = ~mismatch_idx_subworst2subbest & ~mismatch_idx_worst2subworst
                    if all_matched_idx.any():
                        better_output = await node.execute_operation(train_set.iloc[all_matched_idx])
                        node_output.cost += better_output.cost
                        best_output = np.array(list(better_output.outputs.values()))
                        serialized_outputs = self._serialize_map_outputs(better_output.outputs)
                        best_output_embeds, embed_cost = await self.agent.create_embedding(serialized_outputs)
                        node_output.cost += embed_cost
                        cos_sim = cosine_similarity(subbest_output_embeds[all_matched_idx], best_output_embeds)
                        mismatch_idx_subbest2best = cos_sim < 0.5
                        mismatch_ratio = mismatch_idx_subbest2best.sum() / all_matched_idx.sum()
                        subbest_output[:, all_matched_idx] = best_output
                        for idx, key in enumerate(better_output.outputs.keys()):
                            better_output.outputs[key] = subbest_output[idx].tolist()
                    else:
                        mismatch_ratio = 0.0
                    if mismatch_idx_worst2subworst.any():
                        matched_ratio = (
                            (~mismatch_idx_subworst2subbest & mismatch_idx_worst2subworst).sum() / mismatch_idx_worst2subworst.sum()
                        )
                    else:
                        matched_ratio = 0.0
                    improve_score = mismatch_ratio * (1 - improve_score_list[-1]) + improve_score_list[-1] - improve_score_list[-2] + matched_ratio
                    improve_score_list.append(improve_score)

                if improve_score_list[-1] - improve_score_list[-2] > self.improve_margin:
                    exec_model = better_model
                    node_output.outputs = better_output.outputs
        
        optimize_end_time = time.time()
        logger.info(f"Physical Plan Optimization Time: {optimize_end_time - optimize_start_time:.4f} sec")

        node.operator.model = exec_model
        rest_output = await node.execute_operation(test_set)
        node_output = concate_output(node_output, rest_output)
        return node_output

    async def _optimize(self, node: LineageNode) -> tuple[pd.DataFrame, float]:
        if node.left_child:
            dataframe_from_left_node, cost_from_left_branch = await self._optimize(node.left_child)
        if node.right_child:
            dataframe_from_right_node, cost_from_right_branch = await self._optimize(node.right_child)

        if node.op_name == "scan":
            output_from_node = await node.run()
            return output_from_node.output, output_from_node.cost
        
        elif node.op_name == "join":
            token_cost = cost_from_left_branch + cost_from_right_branch
            output_from_node = await self.optimize_exec_model(node, [dataframe_from_left_node, dataframe_from_right_node])
            dataframe_from_node = await node.collate_dataframe([dataframe_from_left_node, dataframe_from_right_node], output_from_node)
            token_cost += output_from_node.cost
            return dataframe_from_node, token_cost
        
        else:
            token_cost = cost_from_left_branch
            output_from_node = await self.optimize_exec_model(node, dataframe_from_left_node)
            dataframe_from_node = await node.collate_dataframe(dataframe_from_left_node, output_from_node)
            token_cost += output_from_node.cost
            return dataframe_from_node, token_cost

    def optimize(
        self, 
        plan: LineageNode,
    ):
        execution_start_time = time.time()
        dataframe_from_node, total_token_cost = asyncio.run(self._optimize(plan))
        execution_end_time = time.time()
        execution_time = execution_end_time - execution_start_time
        return dataframe_from_node, total_token_cost, execution_time



# from nirvana.ops.discover import discover_wrapper as discover
from nirvana.ops.filter import filter_wrapper as filter
from nirvana.ops.map import map_wrapper as map
from nirvana.ops.rank import rank_wrapper as rank
from nirvana.ops.reduce import reduce_wrapper as reduce
from nirvana.ops.join import join_wrapper as join


__all__ = [
    # "discover",
    "filter",
    "map",
    "rank",
    "reduce",
    "select",
    "join"
]



from nirvana.optim.optimizer import OptimizeConfig
from nirvana.optim.logical import LogicalOptimizer
from nirvana.optim.physical import PhysicalOptimizer
from nirvana.optim.optimizer import PlanOptimizer


__all__ = [
    "OptimizeConfig",
    "LogicalOptimizer",
    "PhysicalOptimizer",
    "PlanOptimizer",
]



import pytest
import pandas as pd
from unittest.mock import AsyncMock

import nirvana as nv
from nirvana.ops.join import JoinOperation, JoinOpOutputs
from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import FunctionCallTool


@pytest.fixture
def left_dataframe():
    """Create a sample left dataframe for testing."""
    return pd.DataFrame(
        data={"symptom": ["headache", "cough", "fever"]},
        index=[1, 2, 3]
    )


@pytest.fixture
def right_dataframe():
    """Create a sample right dataframe for testing."""
    return pd.DataFrame(
        data={
            "medical_use": [
                "treat mild to moderate pain, painful menstruation, osteoarthritis, dental pain, headaches, and pain from kidney stones",
                "treat bronchospasm, as well as chronic obstructive pulmonary disease",
                "reduce fever and inflammation"
            ],
        },
        index=[1, 4, 7]
    )


@pytest.fixture
def left_empty_dataframe():
    """Create an empty left dataframe for testing."""
    return pd.DataFrame({"symptom": []})


@pytest.fixture
def right_empty_dataframe():
    """Create an empty right dataframe for testing."""
    return pd.DataFrame({"medical_use": []})


@pytest.fixture
def dataframe_with_nan():
    """Create a dataframe with NaN values for testing."""
    return pd.DataFrame(
        data={"symptom": [None, "cough", "fever"]},
        index=[1, 2, 3]
    )


@pytest.fixture
def mock_llm_client():
    """Create a mock LLM client for testing."""
    
    async def mock_call(messages, parse_tags=False, parse_code=False, **kwargs):
        """Mock LLM call that returns structured output for join operations."""
        result = {
            "raw_output": "<output>False</output>",
            "cost": 0.01
        }
        
        if parse_tags:
            tags = kwargs.get("tags", [])
            # When parse_tags is True, the LLM client extracts XML tags
            for tag in tags:
                if tag == "output":
                    # Return False for join operations by default
                    result[tag] = "False"
        return result
    
    # Create a proper async mock
    mock_client = AsyncMock(spec=LLMClient, side_effect=mock_call)
    mock_client.default_model = "gpt-4"
    return mock_client


@pytest.fixture
def mock_block_join_llm_client():
    """ Mock LLM to return batch join results."""
    async def mock_call_block(messages, parse_tags=False, **kwargs):
        result = {
            "raw_output": "<output></output>",
            "cost": 0.01
        }
        if parse_tags:
            tags = kwargs.get("tags", [])
            for tag in tags:
                if tag == "output":
                    result[tag] = ""
        return result
    
    # Create a proper async mock
    mock_client = AsyncMock(spec=LLMClient, side_effect=mock_call_block)
    mock_client.default_model = "gpt-4"
    return mock_client


class TestJoinOperation:
    """Test suite for JoinOperation class."""
    
    def test_join_operation_initialization(self, mock_llm_client):
        """Test that JoinOperation initializes correctly."""
        JoinOperation.set_llm(mock_llm_client)
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest"
        )
        
        assert op.op_name == "join"
        assert op.left_on == ["symptom"]
        assert op.right_on == ["medical_use"]
        assert op.how == "inner"
        assert op.dependencies == ["symptom", "medical_use"]
        assert op.generated_fields == []
    
    def test_op_kwargs_property(self, mock_llm_client):
        """Test that op_kwargs includes left_on, right_on, and how."""
        JoinOperation.set_llm(mock_llm_client)
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner"
        )
        
        kwargs = op.op_kwargs
        assert "left_on" in kwargs
        assert "right_on" in kwargs
        assert "how" in kwargs
        assert kwargs["left_on"] == ["symptom"]
        assert kwargs["right_on"] == ["medical_use"]
        assert kwargs["how"] == "inner"
    
    @pytest.mark.asyncio
    async def test_execute_with_empty_dataframes(self, left_empty_dataframe, right_empty_dataframe, mock_llm_client):
        """Test that execute returns empty results for empty dataframes."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner"
        )
        
        result = await op.execute(left_data=left_empty_dataframe, right_data=right_empty_dataframe)
        
        assert isinstance(result, JoinOpOutputs)
        # JoinOpOutputs inherits 'output' from BaseOpOutputs
        # The code returns output=[] for empty case
        assert result.output == []
        assert result.left_join_keys == []
        assert result.right_join_keys == []
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_with_nest_strategy(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test execute with nested join strategy."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest"
        )
        cache = {(1, 1): True, (3, 7): True}
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
        
        assert isinstance(result, JoinOpOutputs)
        # The code returns output=joined_pairs, so result.output contains the joined pairs
        assert result.join_pairs == [(1, 1), (3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 2]
        assert result.cost == 0.07
    
    @pytest.mark.asyncio
    async def test_execute_with_block_strategy(self, left_dataframe, right_dataframe, mock_block_join_llm_client):
        """Test execute with block join strategy."""
        JoinOperation.set_llm(mock_block_join_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="block",
            batch_size=2,
        )
        cache = {(0, 0): [(1, 1)], (1, 1): [(3, 7)]}
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
        
        assert result.join_pairs == [(1, 1), (3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 2]
        assert result.cost == 0.02
    
    @pytest.mark.asyncio
    async def test_execute_with_nan_values(self, dataframe_with_nan, right_dataframe, mock_llm_client):
        """Test that execute handles NaN values correctly."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest"
        )
        cache = {(3, 7): True}
        result = await op.execute(left_data=dataframe_with_nan, right_data=right_dataframe, cache=cache)
        
        assert result.join_pairs == [(3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [3, 4, 2]
    
    @pytest.mark.asyncio
    async def test_execute_with_udf(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test execute with user-defined function."""
        JoinOperation.set_llm(mock_llm_client)
        
        def check_match(symptom: pd.Series, medical_use: pd.Series):
            """Simple function to check if symptom matches medical use."""
            if symptom["symptom"] is None or medical_use["medical_use"] is None:
                return False
            return symptom["symptom"].lower() in medical_use["medical_use"].lower()
        
        tool = FunctionCallTool.from_function(func=check_match)
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest",
            tool=tool
        )
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe)
        
        assert result.join_pairs == [(1, 1), (3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 2]
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_with_udf_exception_fallback(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test that UDF exceptions fall back to LLM."""
        JoinOperation.set_llm(mock_llm_client)
        
        def failing_func(left, right):
            """Function that raises an exception."""
            raise ValueError("Function failed")
        
        tool = FunctionCallTool.from_function(func=failing_func)
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest",
            tool=tool
        )
        
        cache = {(1, 1): True, (3, 7): True}
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
        
        assert isinstance(result, JoinOpOutputs)
        # The code returns output=joined_pairs, so result.output contains the joined pairs
        assert result.join_pairs == [(1, 1), (3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 2]
        assert result.cost == 0.07
    
    @pytest.mark.asyncio
    async def test_execute_with_different_join_types(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test execute with different join types (inner, left, right)."""
        JoinOperation.set_llm(mock_llm_client)
        
        for join_type in ["inner", "left", "right"]:
            op = JoinOperation(
                user_instruction="Does the drug cure the symptom?",
                left_on=["symptom"],
                right_on=["medical_use"],
                how=join_type,
                strategy="nest"
            )
            cache = {(1, 1): True, (3, 7): True}
            result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
            
            if join_type in ["inner", "left"]:
                assert result.join_pairs == [(1, 1), (3, 7)]
                assert result.left_join_keys == [0, 1, 2]
                assert result.right_join_keys == [0, 4, 2]
            else:
                assert result.join_pairs == [(1, 1), (3, 7)]
                assert result.left_join_keys == [3, 1, 5]
                assert result.right_join_keys == [3, 4, 5]
    
    @pytest.mark.asyncio
    async def test_execute_unsupported_strategy(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test that execute raises ValueError for unsupported strategy."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="unsupported"
        )
        with pytest.raises(ValueError, match="Strategy unsupported is not supported"):
            await op.execute(left_data=left_dataframe, right_data=right_dataframe)
    
    @pytest.mark.asyncio
    async def test_block_join_with_udf_warning(self, left_dataframe, right_dataframe, mock_block_join_llm_client):
        """Test that block join warns when UDF is provided."""
        import warnings
        
        JoinOperation.set_llm(mock_block_join_llm_client)
        
        def check_match(symptom, medical_use):
            return True
        
        tool = FunctionCallTool.from_function(func=check_match)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="block",
            tool=tool
        )
        
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = await op.execute(left_data=left_dataframe, right_data=right_dataframe)
            
            # Check that warning was issued
            assert len(w) > 0
            assert "block semantic join does not support user-defined functions" in str(w[0].message).lower()

    @pytest.mark.asyncio
    async def test_nested_join_with_limit(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test execute with nested join strategy."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest",
            limit=1,
        )
        cache = {(1, 1): True, (3, 7): True}
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
        
        assert result.join_pairs == [(1, 1)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 5]

    @pytest.mark.asyncio
    async def test_block_join_with_limit(self, left_dataframe, right_dataframe, mock_block_join_llm_client):
        """Test execute with block join strategy."""
        JoinOperation.set_llm(mock_block_join_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="block",
            limit=1,
            batch_size=2,
        )
        cache = {(0, 0): [(1, 1)], (1, 1): [(3, 7)]}
        result = await op.execute(left_data=left_dataframe, right_data=right_dataframe, cache=cache)
        
        assert result.join_pairs == [(1, 1)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 5]


class TestJoinOpOutputs:
    """Test suite for JoinOpOutputs class."""
    
    def test_join_op_outputs_initialization(self):
        """Test that JoinOpOutputs initializes correctly."""
        outputs = JoinOpOutputs(
            join_pairs=[(0, 1), (1, 2)],
            left_join_keys=[0, 1],
            right_join_keys=[1, 2],
            cost=0.05
        )
        
        assert outputs.join_pairs == [(0, 1), (1, 2)]
        assert outputs.left_join_keys == [0, 1]
        assert outputs.right_join_keys == [1, 2]
        assert outputs.cost == 0.05
    
    def test_join_op_outputs_default_values(self):
        """Test that JoinOpOutputs has correct default values."""
        outputs = JoinOpOutputs()
        
        assert outputs.join_pairs == []
        assert outputs.left_join_keys == []
        assert outputs.right_join_keys == []
        assert outputs.cost == 0.0


class TestJoinWrapper:
    """Test suite for join_wrapper function."""
    
    def test_join_wrapper_basic(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test the join_wrapper function with basic usage."""
        JoinOperation.set_llm(mock_llm_client)
        
        cache = {(1, 1): True, (3, 7): True}
        result = nv.ops.join(
            left_data=left_dataframe,
            right_data=right_dataframe,
            user_instruction="Does the drug cure the symptom?",
            left_on="symptom",
            right_on="medical_use",
            how="inner",
            strategy="nest",
            cache=cache
        )
        
        assert isinstance(result, JoinOpOutputs)
        # The code returns output=joined_pairs, so result.output contains the joined pairsassert result.output == [(1, 1), (3, 7)]
        assert result.join_pairs == [(1, 1), (3, 7)]
        assert result.left_join_keys == [0, 1, 2]
        assert result.right_join_keys == [0, 4, 2]
        assert result.cost == 0.07
    
    def test_join_wrapper_with_different_join_types(self, left_dataframe, right_dataframe, mock_llm_client):
        """Test join_wrapper with different join types."""
        JoinOperation.set_llm(mock_llm_client)
        
        cache = {(1, 1): True, (3, 7): True}
        for join_type in ["inner", "left", "right"]:
            result = nv.ops.join(
                left_data=left_dataframe,
                right_data=right_dataframe,
                user_instruction="Does the drug cure the symptom?",
                left_on="symptom",
                right_on="medical_use",
                how=join_type,
                strategy="nest",
                cache=cache
            )

            if join_type in ["inner", "left"]:
                assert result.join_pairs == [(1, 1), (3, 7)]
                assert result.left_join_keys == [0, 1, 2]
                assert result.right_join_keys == [0, 4, 2]
            else:
                assert result.join_pairs == [(1, 1), (3, 7)]
                assert result.left_join_keys == [3, 1, 5]
                assert result.right_join_keys == [3, 4, 5]


class TestJoinOperationIntegration:
    """Integration tests for join operation with real-like scenarios."""
    
    def test_nested_join_postprocess_outputs(self, mock_llm_client):
        """Test that _postprocess_nested_join_outputs handles different output formats."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="nest"
        )
        
        # Simulate results with different output formats
        data_id_pairs = [(0, 0), (0, 2), (1, 0), (1, 2)]
        join_outputs = [True, False, True, False]
        left_keys = pd.Series([0, 1], index=[0, 1])
        right_keys = pd.Series([2, 3], index=[0, 2])
        joined_pairs, left_keys, right_keys = op._postprocess_nested_join_outputs(
            data_id_pairs, join_outputs, "inner", left_keys, right_keys
        )
        
        assert joined_pairs == [(0, 0), (1, 0)]
        assert left_keys == [0, 1]
        assert right_keys == [1, 3]
    
    def test_block_join_postprocess_outputs(self, mock_block_join_llm_client):
        """Test that _postprocess_block_join_outputs handles batch results."""
        JoinOperation.set_llm(mock_block_join_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="block"
        )
        
        # Simulate batch results
        responses = ["L0-R0,L1-R1", "", "", "L0-R0"]
        left_keys_in_batches = [[0, 1], [3]]
        right_keys_in_batches = [[0, 1], [2]]
        batch_ids_pairs = [(0, 0), (0, 1), (1, 0), (1, 1)]
        join_pairs = []
        for idx, batch_pair in enumerate(batch_ids_pairs):
            left_batch_idx, right_batch_idx = batch_pair
            response = responses[idx]
            join_pairs_in_batch = op._process_batch_join_response(
                response, left_keys_in_batches[left_batch_idx], right_keys_in_batches[right_batch_idx]
            )
            join_pairs.extend(join_pairs_in_batch)

        left_join_keys = pd.Series(range(3), index=[0, 1, 3])
        right_join_keys = pd.Series(range(3, 6), index=[0, 1, 2])
        join_pairs, left_keys, right_keys = op._postprocess_block_join_outputs(
            join_pairs, "inner", left_join_keys, right_join_keys
        )
        
        assert join_pairs == [(0, 0), (1, 1), (3, 2)]
        assert left_keys == [0, 1, 2]
        assert right_keys == [0, 1, 2]
    
    def test_prepare_nested_join_pairs(self, mock_llm_client):
        """Test that _prepare_nested_join_pairs creates correct pairs."""
        JoinOperation.set_llm(mock_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner"
        )
        
        left_values = pd.Series(["a", "b"], index=[0, 1])
        right_values = pd.Series(["x", "y"], index=[2, 3])
        
        data_id_pairs = op._prepare_nested_join_pairs(left_values, right_values)
        
        assert len(data_id_pairs) == 4  # 2x2 = 4 pairs
        assert data_id_pairs == [(0, 2), (0, 3), (1, 2), (1, 3)]
    
    def test_prepare_join_batches(self, mock_block_join_llm_client):
        """Test that _prepare_join_batches creates correct batches."""
        JoinOperation.set_llm(mock_block_join_llm_client)
        
        op = JoinOperation(
            user_instruction="Does the drug cure the symptom?",
            left_on=["symptom"],
            right_on=["medical_use"],
            how="inner",
            strategy="block"
        )
        left_values = pd.DataFrame({"symptom": ["a", "b", "c", "d"]}, index=[0, 1, 2, 3])
        right_values = pd.DataFrame({"medical_use": ["x", "y"]}, index=[4, 5])
        left_batches, left_keys, right_batches, right_keys = op._prepare_join_batches(
            left_values, right_values, batch_size=2
        )

        target_left_batches = [
            pd.DataFrame({"symptom": ["a", "b"]}, index=[0, 1]),
            pd.DataFrame({"symptom": ["c", "d"]}, index=[0, 1])
        ]
        target_right_batches = [
            pd.DataFrame({"medical_use": ["x", "y"]}, index=[0, 1])
        ]
        assert all(left.equals(target) for left, target in zip(left_batches, target_left_batches))
        assert all(right.equals(target) for right, target in zip(right_batches, target_right_batches))
        assert left_keys == [[0, 1], [2, 3]]
        assert right_keys == [[4, 5]]



import pytest
import pandas as pd
from unittest.mock import AsyncMock
import warnings

import nirvana as nv
from nirvana.ops.rank import RankOperation, RankOpOutputs
from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import FunctionCallTool


@pytest.fixture
def sample_dataframe():
    """Create a sample dataframe for testing."""
    return pd.DataFrame({
        "Courses": ["Probability and Random Processes", "History", "Database", "Convex Optimization"]
    })


@pytest.fixture
def empty_dataframe():
    """Create an empty dataframe for testing."""
    return pd.DataFrame({"Courses": []})


@pytest.fixture
def dataframe_with_nan():
    """Create a dataframe with NaN values for testing."""
    return pd.DataFrame({
        "Courses": ["Probability and Random Processes", None, "Convex Optimization"]
    })


@pytest.fixture
def mock_llm_client():
    """Create a mock LLM client for testing."""
    
    async def mock_call(messages, parse_tags=False, parse_code=False, **kwargs):
        """Mock LLM call that returns structured output for rank operations."""
        # For ranking, the LLM should return "1" or "2" indicating which item is more relevant
        # For the test case "The course requires the least math":
        # - "Probability and Random Processes" likely requires more math
        # - "Convex Optimization" likely requires less math (or vice versa)
        # We'll return "2" to indicate item 2 is more relevant (requires less math)
        result = {
            "raw_output": "<output>2</output>",
            "cost": 0.01
        }
        
        if parse_tags:
            tags = kwargs.get("tags", [])
            for tag in tags:
                if tag == "output":
                    # Return "2" by default, meaning item 2 is more relevant
                    result[tag] = "2"
        
        return result
    
    # Create a proper async mock
    mock_client = AsyncMock(spec=LLMClient, side_effect=mock_call)
    mock_client.default_model = "gpt-4"
    return mock_client


class TestRankOperation:
    """Test suite for RankOperation class."""
    
    def test_rank_operation_initialization(self, mock_llm_client):
        """Test that RankOperation initializes correctly."""
        RankOperation.set_llm(mock_llm_client)
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"],
            descend=True
        )
        
        assert op.op_name == "rank"
        assert op.input_columns == ["Courses"]
        assert op.descend == True
        assert op.dependencies == ["Courses"]
        assert op.generated_fields == []
    
    def test_op_kwargs_property(self, mock_llm_client):
        """Test that op_kwargs includes input_columns and descend."""
        RankOperation.set_llm(mock_llm_client)
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"],
            descend=True
        )
        
        kwargs = op.op_kwargs
        assert "input_columns" in kwargs
        assert "descend" in kwargs
        assert kwargs["input_columns"] == ["Courses"]
        assert kwargs["descend"] == True
    
    @pytest.mark.asyncio
    async def test_execute_with_empty_dataframe(self, mock_llm_client):
        """Test that execute returns empty results for empty dataframe."""
        RankOperation.set_llm(mock_llm_client)
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"]
        )
        
        empty_df = pd.DataFrame({"Courses": []})
        result = await op.execute(input_data=empty_df)
        
        assert isinstance(result, RankOpOutputs)
        assert result.ranked_indices == []
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_basic_ranking(self, sample_dataframe, mock_llm_client):
        """Test execute with basic ranking."""
        RankOperation.set_llm(mock_llm_client)
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"],
            descend=True
        )
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        result = await op.execute(input_data=sample_dataframe, cache=cache)
        
        assert isinstance(result, RankOpOutputs)
        assert result.ranked_indices == [1, 2, 0, 3]
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_by_ascending_order(self, sample_dataframe, mock_llm_client):
        """Test rank by ascending order."""
        RankOperation.set_llm(mock_llm_client)
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"],
            descend=False
        )
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        result = await op.execute(input_data=sample_dataframe, cache=cache)
        
        assert isinstance(result, RankOpOutputs)
        assert result.ranked_indices == [3, 0, 2, 1]
        assert result.cost == 0.0
    
    @pytest.mark.asyncio
    async def test_execute_with_udf_warning(self, sample_dataframe, mock_llm_client):
        """Test that UDF triggers a warning."""
        RankOperation.set_llm(mock_llm_client)
        
        def compare_func(item1, item2):
            """Function that compares two items."""
            return "1" if len(item1) > len(item2) else "2"
        
        tool = FunctionCallTool.from_function(func=compare_func)
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"],
            tool=tool
        )
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = await op.execute(input_data=sample_dataframe, cache=cache)
            
            # Check that warning was issued
            assert len(w) > 0
            assert "udf is not supported" in str(w[0].message).lower()
            # Should still return a valid result (falls back to LLM)
            assert isinstance(result, RankOpOutputs)
    
    @pytest.mark.asyncio
    async def test_execute_missing_user_instruction_and_func(self, sample_dataframe, mock_llm_client):
        """Test that execute raises ValueError when both user_instruction and func are missing."""
        RankOperation.set_llm(mock_llm_client)
        
        op = RankOperation(
            user_instruction=None,
            input_columns=["Courses"]
        )
        
        with pytest.raises(ValueError, match="Neither `user_instruction` nor `func` is given"):
            await op.execute(input_data=sample_dataframe)


class TestRankOpOutputs:
    """Test suite for RankOpOutputs class."""
    
    def test_rank_op_outputs_initialization(self):
        """Test that RankOpOutputs initializes correctly."""
        outputs = RankOpOutputs(
            ranking=[1, 2],
            ranked_indices=[0, 1],
            cost=0.05
        )
        
        assert outputs.ranking == [1, 2]
        assert outputs.ranked_indices == [0, 1]
        assert outputs.cost == 0.05
    
    def test_rank_op_outputs_default_values(self):
        """Test that RankOpOutputs has correct default values."""
        outputs = RankOpOutputs()
        
        assert outputs.ranking == []
        assert outputs.ranked_indices == []
        assert outputs.cost == 0.0


class TestRankWrapper:
    """Test suite for rank_wrapper function."""
    
    def test_rank_wrapper_basic(self, sample_dataframe, mock_llm_client):
        """Test the rank_wrapper function with basic usage."""
        RankOperation.set_llm(mock_llm_client)
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        result = nv.ops.rank(
            sample_dataframe,
            "The course requires the least math",
            input_column="Courses",
            descend=True,
            cache=cache
        )
        
        assert isinstance(result, RankOpOutputs)
        assert result.ranked_indices == [1, 2, 0, 3]
        assert result.ranking == [3, 1, 2, 4]
        assert result.cost == 0.0
    
    def test_rank_wrapper_by_ascending_order(self, sample_dataframe, mock_llm_client):
        """Test rank_wrapper by ascending order."""
        RankOperation.set_llm(mock_llm_client)
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        result = nv.ops.rank(
            sample_dataframe,
            "The course requires the least math",
            input_column="Courses",
            descend=False,
            cache=cache
        )
        
        assert isinstance(result, RankOpOutputs)
        assert result.ranked_indices == [3, 0, 2, 1]
        assert result.ranking == [2, 4, 3, 1]
        assert result.cost == 0.0


class TestRankOperationIntegration:
    """Integration tests for rank operation with real-like scenarios."""
    
    @pytest.mark.asyncio
    async def test_compare_by_plain_llm(self, mock_llm_client):
        """Test that _compare_by_plain_llm returns correct format."""
        RankOperation.set_llm(mock_llm_client)
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"]
        )
        
        data1 = "Probability and Random Processes"
        data2 = "Convex Optimization"
        
        output, cost = await op._compare_by_plain_llm(
            data1, data2, "The course requires the least math", "str"
        )
        
        # Output should be "1" or "2"
        assert output in ["1", "2"]
        assert cost > 0
    
    @pytest.mark.asyncio
    async def test_partition_function(self, sample_dataframe, mock_llm_client):
        """Test that partition function works correctly."""
        RankOperation.set_llm(mock_llm_client)
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"]
        )
        
        data = sample_dataframe["Courses"]
        ranking = list(range(len(data)))
        low, high = 0, len(data) - 1
        partition_pos, cost = await op.partition(
            data, ranking, low, high, "The course requires the least math", "str", cache=cache
        )
        
        assert isinstance(partition_pos, int)
        assert partition_pos == 0
        assert cost == 0.0
    
    @pytest.mark.asyncio
    async def test_quick_sort_function(self, sample_dataframe, mock_llm_client):
        """Test that quick_sort function works correctly."""
        RankOperation.set_llm(mock_llm_client)
        cache = {
            (0, 1): "2", (0, 2): "2", (0, 3): "1", (1, 2): "1", (1, 3): "1", (2, 3): "1",
            (1, 0): "1", (2, 0): "1", (3, 0): "2", (2, 1): "2", (3, 1): "2", (3, 2): "2"
        }
        
        op = RankOperation(
            user_instruction="The course requires the least math",
            input_columns=["Courses"]
        )
        
        data = sample_dataframe["Courses"]
        ranking = list(range(len(data)))
        low, high = 0, len(data) - 1
        sorted_ranking, cost = await op.quick_sort(
            data, ranking, low, high, "The course requires the least math", "str", cache=cache
        )
        
        assert sorted_ranking == [1, 2, 0, 3]
        assert cost == 0.0



from transformers import AutoModelForCausalLM, AutoTokenizer


class LocalLMClient:
    tokenizer: AutoTokenizer = None
    model: AutoModelForCausalLM = None
    
    @classmethod
    def configure(cls, model_path):
        tokenizer = AutoTokenizer.from_pretrained(model_path)
        model = AutoModelForCausalLM.from_pretrained(model_path)
        cls.tokenizer = tokenizer
        cls.model = model
        return cls
    
    def __call__(self, *args, **kwargs):
        pass



import functools
import asyncio
from typing import Any, Literal
from dataclasses import dataclass, field
import pandas as pd

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.ops.base import BaseOpOutputs, BaseOperation


@dataclass
class ScanOpOutputs(BaseOpOutputs):
    output: Any = field(default=None)


class ScanOperation(BaseOperation):
    """
    Scan operator: Extract data from a data source or an LLM
    """
    
    def __init__(
            self,
            source: Literal["dataframe", "llm"] = "dataframe",
            output_columns: list[str] = [],
            num_samples: int | None = None,
    ):
        super().__init__(
            op_name="scan",
            user_instruction="",
        )
        self.source = source
        self.output_columns = output_columns
        self.num_samples = num_samples

    @property
    def dependencies(self) -> list[str]:
        return []
    
    @property
    def generated_fields(self) -> list[str]:
        return self.output_columns
    
    @property
    def op_kwargs(self):
        kwargs = super().op_kwargs
        kwargs["source"] = self.source
        kwargs["output_columns"] = self.output_columns
        return kwargs
    
    def set_sample_size(self, sample_size: int | None):
        if sample_size is not None:
            assert sample_size > 0, "Sample size must be positive."
        self.num_samples = sample_size
    
    async def scan_from_llm(self, *args, **kwargs):
        raise NotImplementedError("LLM scan operator is not implemented yet.")

    async def execute(
        self, 
        input_data: pd.DataFrame,
        **kwargs
    ):
        if self.source == "dataframe":
            if self.num_samples is not None:
                num_samples = min(self.num_samples, len(input_data))
                output_records = input_data.iloc[:num_samples]
            else:
                output_records = input_data
            return ScanOpOutputs(
                output=output_records,
                cost=0.0
            )
        elif self.source == "llm":
            raise NotImplementedError("LLM scan operator is not implemented yet.")



import asyncio
import pandas as pd

from nirvana.executors.llm_backbone import LLMClient
from nirvana.optim.optimize_prompt import RESULT_EVALUATE_PROMPT


def table_serialize(df: pd.DataFrame) -> str:
    if df.empty:
        return "No data in the output."
    return df.to_json(orient="records", lines=True).strip()


class Evaluator:
    @staticmethod
    async def evaluate(ground_truth: pd.DataFrame, process_result: pd.DataFrame, evaluator: LLMClient):
        if not ground_truth.columns.equals(process_result.columns):
            return 0.0
        
        if ground_truth.size != process_result.size:
            return 0.0
        
        evaluate_prompt = [{
            "role": "user",
            "content": RESULT_EVALUATE_PROMPT.format(
                ground_truth=table_serialize(ground_truth),
                result=table_serialize(process_result)
            )
        }]
        response = await evaluator(messages=evaluate_prompt, parse_tags=True, tags=["score"])
        rating = float(response["score"]) / 10.0
        return rating



import warnings

import asyncio
from typing import Any, Callable, Literal
from dataclasses import dataclass, field
import pandas as pd

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.executors.tools import BaseTool, FunctionCallTool
from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.rank_prompter import RankPrompter


def rank_wrapper(
    input_data: pd.DataFrame,
    user_instruction: str = None,
    input_column: str = None,
    descend: bool = True,
    context: list[dict] | str | None = None,
    model: str | None = None,
    func: Callable = None,
    strategy: Literal["plain"] = "plain",
    rate_limit: int = 16,
    assertions: list[Callable] | None = [],
    **kwargs
):
    """
    A function wrapper for rank operation

    Args:
        input_data (pd.DataFrame): Input dataframe
        user_instruction (str, optional): User instruction. Defaults to None.
        input_column (str, optional): Input column. Defaults to None.
        descend (bool, optional): Whether to rank in descending order (True) or ascending order (False). Defaults to True.
        context (list[dict] | str, optional): Context. Defaults to None.
        model (str, optional): Model. Defaults to None.
        func (Callable, optional): User function. Defaults to None.
        strategy (Literal["plain"], optional): Strategy. Defaults to "plain".
        rate_limit (int, optional): Rate limit. Defaults to 16.
        assertions (list[Callable], optional): Assertions. Defaults to [].
        **kwargs: Additional keyword arguments for OpenAI Clent.
    """

    rank_op = RankOperation(
        user_instruction=user_instruction,
        input_columns=[input_column],
        descend=descend,
        context=context,
        model=model,
        tool=FunctionCallTool.from_function(func=func) if func else None,
        strategy=strategy,
        rate_limit=rate_limit,
        assertions=assertions
    )
    outputs = asyncio.run(rank_op.execute(
        input_data=input_data,
        **kwargs
    ))
    ranking = sorted(range(1, len(outputs.ranked_indices) + 1), key=lambda k: outputs.ranked_indices[k - 1])
    outputs.ranking = ranking
    return outputs


@dataclass
class RankOpOutputs(BaseOpOutputs):
    ranking: list[int] = field(default_factory=list)
    ranked_indices: list[int] = field(default_factory=list)


class RankOperation(BaseOperation):
    """
    RankOperation ranks the rows of a DataFrame column according to a user-specified natural language instruction.
    """
    
    def __init__(
        self,
        user_instruction: str = "",
        input_columns: list[str] = [],
        descend: bool = True,
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: Callable | BaseTool | None = None,
        strategy: Literal["plain"] = "plain",
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
    ):
        if tool and not isinstance(tool, BaseTool):
            tool = FunctionCallTool.from_function(func=tool)
        
        super().__init__(
            op_name="rank",
            user_instruction=user_instruction,
            context=context,
            model=model,
            tool=tool,
            strategy=strategy,  
            rate_limit=rate_limit,
            assertions=assertions,
        )
        self.prompter = RankPrompter()
        self.input_columns = input_columns
        self.descend = descend

    @property
    def dependencies(self) -> list[str]:
        return self.input_columns
    
    @property
    def generated_fields(self) -> list[str]:
        return []
    
    @property
    def op_kwargs(self) -> dict:
        kwargs = super().op_kwargs
        kwargs["input_columns"] = self.input_columns
        kwargs["descend"] = self.descend
        return kwargs

    async def _compare_by_plain_llm(self, data1: Any, data2: Any, user_instruction: str, dtype: str, **kwargs):
        async with self.semaphore:
            if dtype == "text":
                data1 = self.input_columns[0] + ": " + str(data1)
                data2 = self.input_columns[0] + ": " + str(data2)
            full_prompt = self.prompter.generate_prompt(data1, data2, user_instruction, dtype)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            return output["output"], output["cost"]
        
    async def partition(self, data: pd.Series, ranking: list, low: int, high: int, user_instruction: str, dtype: str, **kwargs):
        cache = kwargs.get("cache", None)
        # Choose a pivot element
        pivot_index = ranking[(low + high) // 2]
        pivot_value = data.iloc[pivot_index]
        i = low - 1
        j = high + 1

        cost_for_partiton = 0.0
        while True:
            # Find an element on the left that is smaller than the pivot (less satisfy the instruction)
            while True:
                i += 1
                index_in_data = ranking[i]
                if index_in_data == pivot_index:
                    break
                if cache is not None and (index_in_data, pivot_index) in cache:
                    comparison_result = cache[(index_in_data, pivot_index)]
                    cost_per_comparison = 0.0
                else:
                    comparison_result, cost_per_comparison = await self._compare_by_plain_llm(data.iloc[index_in_data], pivot_value, user_instruction, dtype, **kwargs)
                cost_for_partiton += cost_per_comparison
                if int(comparison_result) == 2:  # if data[i] is smaller than pivot_value, break
                    break

            # Find an element on the right that is larger than the pivot (more satisfy the instruction)
            while True:
                j -= 1
                index_in_data = ranking[j]
                if index_in_data == pivot_index:
                    break
                if cache is not None and (index_in_data, pivot_index) in cache:
                    comparison_result = cache[(index_in_data, pivot_index)]
                    cost_per_comparison = 0.0
                else:
                    comparison_result, cost_per_comparison = await self._compare_by_plain_llm(data.iloc[index_in_data], pivot_value, user_instruction, dtype, **kwargs)
                cost_for_partiton += cost_per_comparison
                if int(comparison_result) == 1:  # if data[j] is larger than pivot_value, break
                    break

            # If i and j have crossed, break
            if i >= j:
                break
            # Swap data[i] and data[j]
            ranking[i], ranking[j] = ranking[j], ranking[i]

        return j, cost_for_partiton
    
    async def quick_sort(self, data: pd.Series, ranking: list, low: int, high: int, user_instruction: str, dtype: str, **kwargs):
        total_cost = 0.0
        if low < high:
            partition_position, cost_for_partition = await self.partition(data, ranking, low, high, user_instruction, dtype, **kwargs)
            total_cost += cost_for_partition
            ranking, cost_from_next_sort = await self.quick_sort(data, ranking, low, partition_position, user_instruction, dtype, **kwargs)
            total_cost += cost_from_next_sort
            ranking, cost_from_next_sort = await self.quick_sort(data, ranking, partition_position + 1, high, user_instruction, dtype, **kwargs)
            total_cost += cost_from_next_sort
        return ranking, total_cost

    async def execute(
        self,
        input_data: pd.DataFrame,
        **kwargs,
    ):
        if self.user_instruction is None and not self.has_udf():
            raise ValueError("Neither `user_instruction` nor `func` is given.")
        
        if input_data.empty:
            return RankOpOutputs(ranked_indices=[], cost=0.0)

        processed_data = input_data[self.input_columns[0]]
        if isinstance(processed_data.dtype, ImageDtype):
            dtype = "image"
        else:
            dtype = "text"

        if self.has_udf():
            warnings.warn("The udf is not supported in the current Rank operator implementation. Switch to LLM-based ranking.")

        low, high = 0, len(processed_data) - 1
        ranking = list(range(len(processed_data)))
        ranking, token_cost = await self.quick_sort(processed_data, ranking, low, high, self.user_instruction, dtype, model=self.model, **kwargs)
        if not self.descend:
            ranking.reverse()

        return RankOpOutputs(
            ranked_indices=ranking,
            cost=token_cost
        )



import warnings

import asyncio
import functools
import pandas as pd
from typing import Any, Iterable, Callable, Literal
from dataclasses import dataclass, field

from nirvana.dataframe.arrays.image import ImageDtype
from nirvana.executors.tools import BaseTool, FunctionCallTool
from nirvana.ops.base import BaseOpOutputs, BaseOperation
from nirvana.ops.prompt_templates.join_prompter import JoinPrompter


def join_wrapper(
    left_data: pd.DataFrame,
    right_data: pd.DataFrame, 
    user_instruction: str,
    left_on: str,
    right_on: str,
    how: str = "inner",
    context: list[dict] | str | None = None,
    model: str | None = None,
    func: Callable = None,
    strategy: Literal["nest", "block"] = "nest",
    limit: int | None = None,
    rate_limit: int = 16,
    assertions: list[Callable] | None = [],
    batch_size: int = 5,
    **kwargs
):
    """
    A function wrapper for join operation

    Args:
        left_data (pd.DataFrame): Left dataframe
        right_data (pd.DataFrame): Right dataframe
        user_instruction (str): User instruction
        left_on (str): Left on
        right_on (str): Right on
        how (str, optional): How. Defaults to "inner".
        context (list[dict] | str, optional): Context. Defaults to None.
        model (str, optional): Model. Defaults to None.
        func (Callable, optional): User function. Defaults to None.
        strategy (Literal["nest", "block"], optional): Strategy. Defaults to "nest".
        limit (int): Maximum number of outputs to produce before stopping.
        rate_limit (int, optional): Rate limit. Defaults to 16.
        assertions (list[Callable], optional): Assertions. Defaults to [].
        batch_size (int, optional): Batch size for block join. Defaults to 5.
        **kwargs: Additional keyword arguments for OpenAI Clent.
    """
    
    join_op = JoinOperation(
        user_instruction=user_instruction,
        left_on=[left_on],
        right_on=[right_on],
        how=how,
        context=context,
        model=model,
        tool=FunctionCallTool(func=func) if func else None,
        strategy=strategy,
        limit=limit,
        rate_limit=rate_limit,
        assertions=assertions,
        batch_size=batch_size,
    )
    outputs = asyncio.run(join_op.execute(
        left_data=left_data,
        right_data=right_data,
        **kwargs
    ))
    return outputs


@dataclass
class JoinOpOutputs(BaseOpOutputs):
    join_pairs: list[tuple] = field(default_factory=list)
    left_join_keys: list[int] = field(default_factory=list)
    right_join_keys: list[int] = field(default_factory=list)


class JoinOperation(BaseOperation):
    """
    Join operator: Join values of two columns against a specific user's instruction.
    """
    strategy_options = ["nest", "block"]
    
    def __init__(
        self,
        user_instruction: str = "",
        left_on: list[str] = [],
        right_on: list[str] = [],
        how: str = "inner",
        context: list[dict] | str | None = None,
        model: str | None = None,
        tool: Callable | BaseTool | None = None,
        strategy: Literal["nest", "block"] = "nest",
        limit: int | None = None,
        rate_limit: int = 16,
        assertions: list[Callable] | None = [],
        batch_size: int = 5,
    ):
        if tool and not isinstance(tool, BaseTool):
            tool = FunctionCallTool.from_function(func=tool)
        
        super().__init__(
            op_name="join", 
            user_instruction=user_instruction,
            context=context,
            model=model,
            tool=tool,
            strategy=strategy,
            limit=limit,
            rate_limit=rate_limit,
            assertions=assertions,
        )
        self.prompter = JoinPrompter()
        self.left_on = left_on
        self.right_on = right_on
        self.how = how
        self.batch_size = batch_size

    @property
    def dependencies(self) -> list[str]:
        return self.left_on + self.right_on

    @property
    def generated_fields(self) -> list[str]:
        return []
    
    @property
    def op_kwargs(self) -> dict:
        kwargs = super().op_kwargs
        kwargs["left_on"] = self.left_on
        kwargs["right_on"] = self.right_on
        kwargs["how"] = self.how
        if self.strategy == "block":
            kwargs["batch_size"] = self.batch_size
        return kwargs
    
    def _prepare_nested_join_pairs(self, left_values, right_values):
        left_ids = left_values.index
        right_ids = right_values.index
        data_id_pairs = [(left_id, right_id) for left_id in left_ids for right_id in right_ids]
        return data_id_pairs
    
    async def _execute_by_func(self, left_value: pd.Series, right_value: pd.Series, user_instruction: str, func: Callable, llm_call: Callable, **kwargs):
        try:
            result = func(left_value, right_value)
            return result, 0.0
        except Exception as e:
            warnings.warn(f"Evaluation by UDF failed with error {e}. Switch to LLM evaluation.")
            return await llm_call(left_value, right_value, user_instruction)

    async def _pairwise_evaluate(self, left_value: pd.Series, right_value: pd.Series, user_instruction: str, left_dtypes: list[str], right_dtypes: list[str], **kwargs):
        async with self.semaphore:
            full_prompt = self.prompter.generate_nested_join_prompt(left_value, right_value, user_instruction, left_dtypes, right_dtypes)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            result = self._process_nested_join_response(output.get("output", None))
            return result, output["cost"]
        
    def _process_nested_join_response(self, response: str | None) -> bool:
        if response is None:
            return False
        elif isinstance(response, bool):
            return response
        elif "true" in response.lower():
            return True
        elif "false" in response.lower():
            return False
        else:
            return False
    
    def _postprocess_nested_join_outputs(
        self, 
        data_id_pairs: list[tuple], 
        join_outputs: list[bool],
        how: str,
        left_keys: pd.Series,
        right_keys: pd.Series,
    ) -> tuple[list[tuple], list[int], list[int]]:
        key_mapping = {}
        join_pairs = []
        for can_join, (left_key, right_key) in zip(join_outputs, data_id_pairs):
            if can_join:
                join_pairs.append((left_key, right_key))
                # If left join, keys for right side are mapped to left side;
                # otherwise, keys for left side are mapped to right side.
                if how == "inner" or how == "left":
                    key_mapping[right_keys.loc[right_key]] = left_keys.loc[left_key]
                else:
                    key_mapping[left_keys.loc[left_key]] = right_keys.loc[right_key]
        
        if how == "inner" or how == "left":
            right_join_keys = right_keys.apply(lambda x: key_mapping.get(x, x)).to_list()
            left_join_keys = left_keys.to_list()
        else:
            left_join_keys = left_keys.apply(lambda x: key_mapping.get(x, x)).to_list()
            right_join_keys = right_keys.to_list()
        return join_pairs, left_join_keys, right_join_keys
    
    async def _nested_join(
        self, 
        left_data: pd.DataFrame, 
        right_data: pd.DataFrame, 
        user_instruction: str, 
        left_dtypes: list[str], 
        right_dtypes: list[str],
        **kwargs
    ):
        cache = kwargs.pop("cache", None)
        execution_func = functools.partial(self._pairwise_evaluate, left_dtypes=left_dtypes, right_dtypes=right_dtypes, model=self.model, **kwargs)
        # Prepare candidate pairs
        data_id_pairs = self._prepare_nested_join_pairs(left_data, right_data)
        left_keys = pd.Series(range(len(left_data)), index=left_data.index)
        right_keys = pd.Series(range(len(left_data), len(left_data) + len(right_data)), index=right_data.index)
        
        # Prepare values for join and their dtypes
        left_join_values: pd.DataFrame = left_data[self.left_on]
        right_join_values: pd.DataFrame = right_data[self.right_on]

        tasks = []
        for left_id, right_id in data_id_pairs:
            left_value: pd.Series = left_join_values.loc[left_id]
            right_value: pd.Series = right_join_values.loc[right_id]
            if left_value.empty or right_value.empty:
                tasks.append(asyncio.create_task(asyncio.sleep(0, result=(False, 0.0))))
            elif cache is not None and (left_id, right_id) in cache:
                join_result = (cache[(left_id, right_id)], 0.0)
                tasks.append(asyncio.create_task(asyncio.sleep(0, result=join_result)))
            elif self.has_udf():
                tasks.append(asyncio.create_task(self._execute_by_func(left_value, right_value, user_instruction, self.tool, execution_func)))
            else:
                tasks.append(asyncio.create_task(execution_func(left_value, right_value, user_instruction)))
        
        # Wait for all tasks to complete
        if self.limit is not None and self.limit <= 0:
            warnings.warn("The limit should be positive. To execute, the limit will be ignored.")
            self.limit = None
        
        token_cost = 0.0
        join_outputs: list[bool] = []
        if self.limit is not None:
            num_passed_pairs: int = 0
            reach_limit: bool = False
            for i in range(0, len(tasks), self.limit):
                if reach_limit:
                    break
                batch_tasks = tasks[i : i + self.limit]
                batch_results = await asyncio.gather(*batch_tasks)
                token_cost += sum(result[1] for result in batch_results)
                for can_join, _ in batch_results:
                    join_outputs.append(can_join)
                    if can_join:
                        num_passed_pairs += 1
                    if num_passed_pairs >= self.limit:
                        reach_limit = True
                        break
            num_remaining_pairs = len(data_id_pairs) - len(join_outputs)
            if num_remaining_pairs > 0:
                join_outputs.extend([False] * num_remaining_pairs)
        else:
            results = await asyncio.gather(*tasks)
            join_outputs = [result[0] for result in results]
            token_cost = sum(result[1] for result in results)

        joined_pairs, left_join_keys, right_join_keys = self._postprocess_nested_join_outputs(
            data_id_pairs, join_outputs, self.how, left_keys=left_keys, right_keys=right_keys
        )
        return JoinOpOutputs(
            join_pairs=joined_pairs,
            left_join_keys=left_join_keys,
            right_join_keys=right_join_keys,
            cost=token_cost,
        )
    
    def _prepare_join_batches(
        self, 
        left_values: pd.DataFrame,
        right_values: pd.DataFrame,
        batch_size: int,
    ) -> tuple[list[pd.DataFrame], list[list], list[pd.DataFrame], list[list]]:
        # Prepare left batches
        left_batches, left_keys = [], []
        start_idx = 0
        while start_idx < len(left_values):
            left_batches.append(left_values.iloc[start_idx : start_idx+batch_size].reset_index(drop=True))
            left_keys.append(left_values.index[start_idx : start_idx+batch_size].tolist())
            start_idx += batch_size
        
        # Prepare right batches
        right_batches, right_keys = [], []
        start_idx = 0
        while start_idx < len(right_values):
            right_batches.append(right_values.iloc[start_idx : start_idx+batch_size].reset_index(drop=True))
            right_keys.append(right_values.index[start_idx : start_idx+batch_size].tolist())
            start_idx += batch_size
        
        return left_batches, left_keys, right_batches, right_keys

    async def _batchwise_evaluate(
        self,
        left_batch: pd.DataFrame,
        right_batch: pd.DataFrame,
        user_instruction: str,
        left_dtypes: list[str],
        right_dtypes: list[str],
        keys_in_left_batch: list,
        keys_in_right_batch: list,
        **kwargs
    ):
        async with self.semaphore:
            full_prompt = self.prompter.generate_batch_join_prompt(left_batch, right_batch, user_instruction, left_dtypes, right_dtypes)
            output = await self.llm(full_prompt, parse_tags=True, tags=["output"], **kwargs)
            result = self._process_batch_join_response(output.get("output", None), keys_in_left_batch, keys_in_right_batch)
            return result, output["cost"]
        
    def _process_batch_join_response(
        self,
        response: str | None,
        keys_in_left_batch: list,
        keys_in_right_batch: list,
    ) -> list[tuple]:
        if response is None or response == "":
            return []
        pairs_str = response.split(',')
        joined_pairs = []
        for pair_str in pairs_str:
            pair_str = pair_str.strip()
            left_ref, right_ref = pair_str.split('-')
            left_idx, right_idx = keys_in_left_batch[int(left_ref[1:])], keys_in_right_batch[int(right_ref[1:])]
            pair = (left_idx, right_idx)
            joined_pairs.append(pair)
        return joined_pairs
        
    def _postprocess_block_join_outputs(
        self,
        join_pairs: list[tuple],
        how: str,
        left_keys: pd.Series,
        right_keys: pd.Series,
    ):
        key_mapping = {}
        for left_key, right_key in join_pairs:
            # If left join, keys for right side are mapped to left side;
            # otherwise, keys for left side are mapped to right side.
            if how == "inner" or how == "left":
                key_mapping[right_keys.loc[right_key]] = left_keys.loc[left_key]
            else:
                key_mapping[left_keys.loc[left_key]] = right_keys.loc[right_key]

        if how == "inner" or how == "left":
            right_join_keys = right_keys.apply(lambda x: key_mapping.get(x, x)).to_list()
            left_join_keys = left_keys.to_list()
        else:
            left_join_keys = left_keys.apply(lambda x: key_mapping.get(x, x)).to_list()
            right_join_keys = right_keys.to_list()
        return join_pairs, left_join_keys, right_join_keys
    
    async def _block_join(
        self,
        left_data: pd.DataFrame,
        right_data: pd.DataFrame,
        user_instruction: str,
        batch_size: int,
        left_dtypes: list[str],
        right_dtypes: list[str],
        **kwargs
    ):
        cache = kwargs.pop("cache", None)
        execution_func = functools.partial(self._batchwise_evaluate, left_dtypes=left_dtypes, right_dtypes=right_dtypes, model=self.model, **kwargs)

        # Prepare batches
        left_join_values: pd.DataFrame = left_data[self.left_on]
        right_join_values: pd.DataFrame = right_data[self.right_on]
        left_batches, left_keys, right_batches, right_keys = self._prepare_join_batches(left_join_values, right_join_values, batch_size=batch_size)

        tasks, batch_ids_pairs = [], []
        for left_batch_id, (left_batch, keys_in_left_batch) in enumerate(zip(left_batches, left_keys)):
            for right_batch_id, (right_batch, keys_in_right_batch) in enumerate(zip(right_batches, right_keys)):
                batch_ids_pairs.append((left_batch_id, right_batch_id))
                if cache is not None and (left_batch_id, right_batch_id) in cache:
                    join_result = (cache[(left_batch_id, right_batch_id)], 0.0)
                    tasks.append(asyncio.create_task(asyncio.sleep(0, result=join_result)))
                else:
                    tasks.append(asyncio.create_task(execution_func(
                        left_batch, right_batch, user_instruction, keys_in_left_batch=keys_in_left_batch, keys_in_right_batch=keys_in_right_batch
                    )))
        
        # Wait for all tasks to complete
        if self.limit is not None and self.limit <= 0:
            warnings.warn("The limit should be positive. To execute, the limit will be ignored.")
            self.limit = None
        
        token_cost = 0.0
        join_outputs: list[tuple] = []
        if self.limit is not None:
            num_passed_pairs: int = 0
            reach_limit: bool = False
            for i in range(0, len(tasks), self.limit):
                if reach_limit:
                    break
                batch_tasks = tasks[i : i + self.limit]
                batch_results = await asyncio.gather(*batch_tasks)
                token_cost += sum(result[1] for result in batch_results)
                for join_pairs_in_batch, _ in batch_results:
                    margin = self.limit - num_passed_pairs
                    if len(join_pairs_in_batch) <= margin:
                        join_outputs.extend(join_pairs_in_batch)
                    else:
                        join_outputs.extend(join_pairs_in_batch[:margin])
                    num_passed_pairs += len(join_pairs_in_batch)
                    if num_passed_pairs >= self.limit:
                        reach_limit = True
                        break
        else:
            results = await asyncio.gather(*tasks)
            for result in results:
                join_outputs.extend(result[0])
            token_cost = sum(result[1] for result in results)

        left_join_keys = pd.Series(range(len(left_data)), index=left_data.index)
        right_join_keys = pd.Series(range(len(left_data), len(left_data) + len(right_data)), index=right_data.index)
        joined_pairs, left_join_keys, right_join_keys = self._postprocess_block_join_outputs(
            join_outputs, self.how, left_join_keys, right_join_keys
        )
        return JoinOpOutputs(
            join_pairs=joined_pairs,
            left_join_keys=left_join_keys,
            right_join_keys=right_join_keys,
            cost=token_cost,
        )

    async def execute(
        self, 
        left_data: pd.DataFrame,
        right_data: pd.DataFrame,
        **kwargs
    ):
        if self.user_instruction is None and not self.has_udf():
            raise ValueError("`user_instruction` or `tool` (e.g., a UDF) is required.")
        if left_data.empty or right_data.empty:
            return JoinOpOutputs(
                output=[],
                left_join_keys=[],
                right_join_keys=[],
                cost=0.0,
            )
        
        # Prepare dtypes for left and right join columns
        left_dtypes: list = []
        for col in self.left_on:
            if isinstance(left_data[col].dtype, ImageDtype):
                left_dtypes.append("image")
            else:
                left_dtypes.append("text")
        right_dtypes: list = []
        for col in self.right_on:
            if isinstance(right_data[col].dtype, ImageDtype):
                right_dtypes.append("image")
            else:
                right_dtypes.append("text")

        if self.strategy == "nest":
            return await self._nested_join(left_data, right_data, self.user_instruction, left_dtypes, right_dtypes, **kwargs)
        elif self.strategy == "block":
            if self.has_udf():
                warnings.warn("The block semantic join does not support user-defined functions for now.")
            return await self._block_join(left_data, right_data, self.user_instruction, self.batch_size, left_dtypes, right_dtypes, **kwargs)
        else:
            raise ValueError(f"The optional strategies available for join are {self.strategy_options}. Strategy {self.strategy} is not supported.")



import pytest
import pandas as pd
from unittest.mock import AsyncMock

import nirvana as nv
from nirvana.ops.filter import FilterOperation, FilterOpOutputs
from nirvana.executors.llm_backbone import LLMClient
from nirvana.executors.tools import FunctionCallTool


@pytest.fixture
def sample_dataframe():
    """Create a sample dataframe for testing."""
    return pd.DataFrame({
        "Courses": [
            "Probability and Random Processes",
            "History",
            "Database",
            "Convex Optimization",
        ]
    })


@pytest.fixture
def empty_dataframe():
    """Create an empty dataframe for testing."""
    return pd.DataFrame({"Courses": []})


@pytest.fixture
def dataframe_with_nan():
    """Create a dataframe with NaN values for testing."""
    return pd.DataFrame({
        "Courses": [
            "Probability and Random Processes",
            None,
            "Convex Optimization",
        ]
    })


@pytest.fixture
def mock_llm_client():
    """Create a mock LLM client for testing filter operations."""

    async def mock_call(messages, parse_tags=False, parse_code=False, **kwargs):
        """Mock LLM call that returns structured boolean output for filter."""
        text = str(messages)
        # Courses that require a lot of math:
        # - "Probability and Random Processes"
        # - "Convex Optimization"
        if "Probability and Random Processes" in text or "Convex Optimization" in text:
            value = "True"
        elif "None" in text:
            value = "False"
        else:
            value = "False"

        result = {
            "raw_output": f"<output>{value}</output>",
            "cost": 0.01,
        }

        if parse_tags:
            tags = kwargs.get("tags", [])
            for tag in tags:
                if tag == "output":
                    result[tag] = value

        return result

    mock_client = AsyncMock(spec=LLMClient, side_effect=mock_call)
    mock_client.default_model = "gpt-4"
    return mock_client


class TestFilterOperation:
    """Test suite for FilterOperation class."""

    def test_filter_operation_initialization(self, mock_llm_client):
        """Test that FilterOperation initializes correctly."""
        FilterOperation.set_llm(mock_llm_client)
        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        assert op.op_name == "filter"
        assert op.input_columns == ["Courses"]
        assert op.dependencies == ["Courses"]
        assert op.generated_fields == []

    def test_op_kwargs_property(self, mock_llm_client):
        """Test that op_kwargs includes input_columns."""
        FilterOperation.set_llm(mock_llm_client)
        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        kwargs = op.op_kwargs
        assert "input_columns" in kwargs
        assert kwargs["input_columns"] == ["Courses"]

    @pytest.mark.asyncio
    async def test_execute_with_empty_dataframe(self, mock_llm_client):
        """Test that execute returns default outputs for empty dataframe."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
        )

        empty_df = pd.DataFrame({"Courses": []})
        result = await op.execute(input_data=empty_df)

        assert isinstance(result, FilterOpOutputs)
        assert result.output is None or result.output == []
        assert result.cost == 0.0

    @pytest.mark.asyncio
    async def test_execute_with_plain_strategy(self, sample_dataframe, mock_llm_client):
        """Test execute with plain LLM strategy."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        result = await op.execute(input_data=sample_dataframe)

        assert isinstance(result, FilterOpOutputs)
        assert result.output == [True, False, False, True]
        assert result.cost > 0

    @pytest.mark.asyncio
    async def test_execute_with_nan_values(self, dataframe_with_nan, mock_llm_client):
        """Test that execute handles NaN values correctly."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        result = await op.execute(input_data=dataframe_with_nan)

        assert isinstance(result, FilterOpOutputs)
        # NaN rows should be treated as False
        assert result.output == [True, False, True]
        assert result.cost > 0

    @pytest.mark.asyncio
    async def test_execute_with_udf(self, sample_dataframe, mock_llm_client):
        """Test execute with user-defined function."""
        FilterOperation.set_llm(mock_llm_client)

        def requires_lot_of_math(course: pd.Series) -> bool:
            """Simple predicate to identify math-heavy courses."""
            return course["Courses"] in {
                "Probability and Random Processes",
                "Convex Optimization",
            }

        tool = FunctionCallTool.from_function(func=requires_lot_of_math)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            tool=tool,
        )

        result = await op.execute(input_data=sample_dataframe)

        assert isinstance(result, FilterOpOutputs)
        assert result.output == [True, False, False, True]
        assert result.cost == 0.0

    @pytest.mark.asyncio
    async def test_execute_with_udf_exception_fallback(self, sample_dataframe, mock_llm_client):
        """Test that UDF exceptions fall back to LLM."""
        FilterOperation.set_llm(mock_llm_client)

        def failing_func(course: str) -> bool:
            """Function that always raises an exception."""
            raise ValueError("Function failed")

        tool = FunctionCallTool.from_function(func=failing_func)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            tool=tool,
            strategy="plain",
        )

        result = await op.execute(input_data=sample_dataframe)

        # Should fall back to LLM, so result should still be valid
        assert isinstance(result, FilterOpOutputs)
        assert result.output == [True, False, False, True]
        assert result.cost > 0

    @pytest.mark.asyncio
    async def test_execute_missing_user_instruction_and_func(self, sample_dataframe, mock_llm_client):
        """Test that execute raises ValueError when both user_instruction and func are missing."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction=None,
            input_columns=["Courses"],
        )

        with pytest.raises(ValueError, match="Neither `user_instruction` nor `func` is given."):
            await op.execute(input_data=sample_dataframe)

    @pytest.mark.asyncio
    async def test_execute_fewshot_without_context(self, sample_dataframe, mock_llm_client):
        """Test that execute raises AssertionError for fewshot without context."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="fewshot",
        )

        with pytest.raises(AssertionError, match="Few-shot examples must be provided"):
            await op.execute(input_data=sample_dataframe)

    @pytest.mark.asyncio
    async def test_filter_operation_with_limit(self, sample_dataframe, mock_llm_client):
        """Test that filter operation respects the limit on passed records."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
            limit=1,
        )

        result = await op.execute(input_data=sample_dataframe)

        assert isinstance(result, FilterOpOutputs)
        # Only one True should be allowed, remaining rows padded with False
        assert result.output[0] is True
        assert all(v is False for v in result.output[1:])

    def test_postprocess_filter_output(self, mock_llm_client):
        """Test that _postprocess_filter_output normalizes various LLM outputs."""
        FilterOperation.set_llm(mock_llm_client)

        op = FilterOperation(
            user_instruction="The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        # Bool inputs are returned as-is
        assert op._postprocess_filter_output(True) is True
        assert op._postprocess_filter_output(False) is False

        # None becomes False
        assert op._postprocess_filter_output(None) is False

        # String inputs with true/false (any case) are handled
        assert op._postprocess_filter_output("True") is True
        assert op._postprocess_filter_output(" FALSE ") is False


class TestFilterOpOutputs:
    """Test suite for FilterOpOutputs class."""

    def test_filter_op_outputs_initialization(self):
        """Test that FilterOpOutputs initializes correctly."""
        outputs = FilterOpOutputs(
            output=[True, False, True],
            cost=0.05,
        )

        assert outputs.output == [True, False, True]
        assert outputs.cost == 0.05

    def test_filter_op_outputs_default_values(self):
        """Test that FilterOpOutputs has correct default values."""
        outputs = FilterOpOutputs()

        assert outputs.output is None
        assert outputs.cost == 0.0

    def test_filter_op_outputs_addition(self):
        """Test that FilterOpOutputs can be added together."""
        outputs1 = FilterOpOutputs(output=[True, False], cost=0.02)
        outputs2 = FilterOpOutputs(output=[False, True], cost=0.03)

        combined = outputs1 + outputs2

        assert combined.output == [True, False, False, True]
        assert combined.cost == 0.05


class TestFilterWrapper:
    """Test suite for filter_wrapper function."""

    def test_filter_wrapper_basic(self, sample_dataframe, mock_llm_client):
        """Test the filter_wrapper function with basic usage."""
        FilterOperation.set_llm(mock_llm_client)

        result = nv.ops.filter(
            sample_dataframe,
            "The course requires a lot of math",
            input_columns=["Courses"],
            strategy="plain",
        )

        assert isinstance(result, FilterOpOutputs)
        assert result.output == [True, False, False, True]
        assert result.cost > 0



import os
from typing import Optional
from collections.abc import MutableSequence
import bm25s
import chromadb


class LakeMetadata:
    path: str
    context: Optional[str] = None
    column_summary: Optional[str] = None
    record_summary: Optional[str] = None
    is_summarized: bool = False
    is_vector_indexed: bool = False
    is_text_indexed: bool = False


class LakeManager(MutableSequence):
    """
    Core internal intermediator for the data lake (e.g., local storage, S3, and more).

    Manage metadata for a data lake, for example, tracking the size of the data lake,
    providing accesses to data files in the data lake, storing data summaries, etc.
    """

    def __init__(self, vector_index_path: str = None, text_index_path: str = None):
        super().__init__()
        self.vector_index_path = vector_index_path
        self.text_index_path = text_index_path
        self.vector_index = self.__init_vector_index__()
        self.text_index = self.__init_text_index__()

    def __init_vector_index__(self):
        client = chromadb.PersistentClient(path=self.vector_base_path)
        vector_db = client.create_collection(
            name="vector_index", 
            metadata={
                "hnsw:space": "cosine",
                "hnsw:random_seed": 42,
                "hnsw:M": 48,
            }
        )
        return vector_db
    
    def __init_text_index__(self):
        indexer = bm25s.BM25(corpus=[])
        corpus_tokens = bm25s.tokenize([])
        indexer.index(corpus_tokens, show_progress=False)
        indexer.save(
            os.path.join(self.text_index_path, "bm25s_index")
        )
        return indexer

    def __setitem__(
            self, 
            index: int, 
            value: LakeMetadata
    ):
        super().__setitem__(index, value)

    def __getitem__(self, index: int) -> LakeMetadata:
        return super().__getitem__(index)
    
    def __len__(self) -> int:
        return super().__len__()
    
    def __delitem__(self, index: int):
        super().__delitem__(index)

    def insert(self, index, value: LakeMetadata):
        return super().insert(index, value)



from typing import Mapping


class Field:
    def __init__(self):
        self._name: str = None
        self._type = None
        self._metadata: Mapping[str, str] = None

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    def from_dict(self, data: Mapping[str, str]):
        self._name = data["name"]
        self._type = data["type"]
        self._metadata = data["metadata"]



import asyncio
from enum import Enum

from nirvana.executors.tools import FunctionCallTool
from nirvana.executors.llm_backbone import LLMClient
from nirvana.lineage.abstractions import LineageNode


class FusionType(Enum):
    SAME_TYPE = 1
    MAP_FILTER = 2


class OperatorFusion:
    """
    Rule-based operator fusion on the lineage tree.

    Two main patterns are supported:
    1) Merge multiple same-type operators (e.g., map-map or filter-filter) that work
       on the same column into a single operator whose schema (e.g., user_instruction)
       preserves all original instructions.
    2) Merge a filter and a map that work on the same column into a single map
       operator whose instruction is rewritten to perform both the filtering
       and mapping; a trailing filter with a UDF is then added to remove None results.
    """

    rewrite_prompt_for_map_filter_fusion = """You are given a filter and a map operator, each with an NL instruction. The two operations are gonna be merged into a single map operator.
To keep semantically equivalent, you are required to give a new instruction for the merged operator merging. The requirement for the new instruction is:
The new instruction follows the logic that if data don't pass the filter condition, return None for each field as the map result; otherwise, perform data transformation as normal.

The operations and their instructions are follows:
{instruction}

Output the merged instruction concisely in the following format.
<instruction> The merged instruction </instruction>
"""

    rewrite_prompt_for_same_type_fusion = """There are two same-type semantic data analytics operations and each process data based on a given instruction. The operations are gonna be merged into a single one.
To keep semantically equivalent, you are required to give a new instruction for the merged operator merging. The requirement for the new instruction is:
The new instruction must preserve all key data processing requirements.

The operations to merge and their instructions are follows:
{instruction}

Output the merged instruction concisely in the following format.
<instruction> The merged instruction </instruction>
"""

    @classmethod
    def match_pattern(cls, node_a: LineageNode, node_b: LineageNode, pattern: FusionType) -> bool:
        if node_a.operator.has_udf() or node_b.operator.has_udf():
            return False
        
        if pattern == FusionType.SAME_TYPE:
            return node_a.op_name == node_b.op_name
        else:
            return (
                (node_a.op_name in {"map", "filter"} and node_b.op_name in {"map", "filter"}) and 
                (node_a.operator.dependencies == node_b.operator.dependencies)
            )

    @classmethod
    async def transform(cls, node: LineageNode, rewriter: LLMClient) -> tuple[LineageNode, float]:
        return await cls.apply(node, rewriter=rewriter)

    @classmethod
    async def apply(cls, node: LineageNode, rewriter: LLMClient) -> tuple[LineageNode, float]:
        if node.op_name == "scan":
            return node, 0.0
        
        if node.op_name == "join":
            left_child, rewrite_cost_from_left_branch = await cls.apply(node.left_child, rewriter)
            right_child, rewrite_cost_from_right_branch = await cls.apply(node.right_child, rewriter)
            rewrite_cost = rewrite_cost_from_left_branch + rewrite_cost_from_right_branch
            if left_child:
                node.set_left_child(left_child)
            if right_child:
                node.set_right_child(right_child)
            return node, rewrite_cost
        
        if node.op_name in {"rank", "reduce"}:
            child, rewrite_cost = await cls.apply(node.left_child, rewriter)
            node.set_left_child(child)
            return node, rewrite_cost

        child, rewrite_cost = await cls.apply(node.left_child, rewriter)
        node.set_left_child(child)
        if cls.match_pattern(node, child, FusionType.SAME_TYPE):
            node, rewrite_cost_for_cur_node = await cls._merge_same_type_operators(node, child, rewriter)
            rewrite_cost += rewrite_cost_for_cur_node
        if cls.match_pattern(node, child, FusionType.MAP_FILTER):
            node, rewrite_cost_for_cur_node = await cls._merge_map_and_filter(node, child, rewriter)
            rewrite_cost += rewrite_cost_for_cur_node
        return node, rewrite_cost
    
    @classmethod
    async def _merge_same_type_operators(
        cls,
        node: LineageNode,
        node_to_merge: LineageNode,
        rewriter: LLMClient,
    ) -> tuple[LineageNode, float]:
        """
        Merge operators that share a same type as `node` and depend on the same column.
        """
        # Prepare new instruction for the fused operator
        combined_instruction = (
            f"{node.op_name} 1: {node.operator.user_instruction}\n"
            f"{node.op_name} 2: {node_to_merge.operator.user_instruction}"
        )
        prompt = cls.rewrite_prompt_for_same_type_fusion.format(instruction=combined_instruction)
        llm_output = await rewriter(prompt, parse_tags=True, tags=["instruction"])
        new_instruction, cost = llm_output["instruction"], llm_output["cost"]

        # Prepare new operator
        op_name = node.op_name
        op_kwargs: dict = node.operator.op_kwargs
        op_kwargs["user_instruction"] = new_instruction
        dependencies = set(node.operator.dependencies + node_to_merge.operator.dependencies)
        op_kwargs["input_columns"] = list(dependencies)

        if op_name == "map":
            generated_fields = set(node.operator.generated_fields + node_to_merge.operator.generated_fields)
            op_kwargs["output_columns"] = list(generated_fields)
        node_fields = {
            "left_input_fields": node_to_merge.node_fields.left_input_fields,
            "right_input_fields": [],
            "output_fields": list(
                set(node_to_merge.node_fields.left_input_fields + op_kwargs.get("output_columns", []))
            )
        }
        new_op = LineageNode(
            op_name=op_name,
            op_kwargs=op_kwargs,
            node_fields=node_fields,
        )
        new_op.set_left_child(node_to_merge.left_child)
        return new_op, cost

    @classmethod
    async def _merge_map_and_filter(
        cls,
        node: LineageNode,
        node_to_merge: LineageNode,
        rewriter: LLMClient,
    ) -> tuple[LineageNode, float]:
        """
        Merge a map and a filter into a single map followed by a code filter
        with a function `lambda x: x is not None`. 
        I.e., `Map + Filter -> Map + Filter(lambda x: x is not None)`
        or `Filter + Map -> Map + Filter(lambda x: x is not None)`
        """
        # Prepare new instruction for the fused operator
        combined_instruction = (
            f"{node.op_name} 1: {node.operator.user_instruction}\n"
            f"{node_to_merge.op_name} 1: {node_to_merge.operator.user_instruction}"
        )
        prompt = cls.rewrite_prompt_for_map_filter_fusion.format(instruction=combined_instruction)
        llm_output = await rewriter(prompt, parse_tags=True, tags=["instruction"])
        new_instruction, cost = llm_output["instruction"], llm_output["cost"]

        # Prepare new map operator
        op_kwargs = node.operator.op_kwargs if node.op_name == "map" else node_to_merge.operator.op_kwargs
        op_kwargs["user_instruction"] = new_instruction
        dependencies = set(node.operator.dependencies + node_to_merge.operator.dependencies)
        op_kwargs["input_columns"] = list(dependencies)

        generated_fields = node.operator.generated_fields if node.op_name == "map" else node_to_merge.operator.generated_fields
        op_kwargs["output_columns"] = list(generated_fields)
        node_fields = {
            "left_input_fields": node_to_merge.node_fields.left_input_fields,
            "right_input_fields": [],
            "output_fields": list(
                set(node_to_merge.node_fields.left_input_fields + op_kwargs.get("output_columns", []))
            )
        }
        new_map_node = LineageNode(
            op_name="map",
            op_kwargs=op_kwargs,
            node_fields=node_fields,
        )
        new_map_node.set_left_child(node_to_merge.left_child)

        filter_kwargs = {
            "user_instruction": f"Fields {op_kwargs['output_columns']} contain None",
            "input_columns": op_kwargs["output_columns"],
            "tool": FunctionCallTool.from_function(func=lambda x: x.hasnans),
        }
        node_fields = {
            "left_input_fields": new_map_node.node_fields.output_fields,
            "right_input_fields": [],
            "output_fields": new_map_node.node_fields.output_fields,
        }
        code_filter = LineageNode(
            op_name="filter",
            op_kwargs=filter_kwargs,
            node_fields=node_fields
        )
        code_filter.set_left_child(new_map_node)
        return code_filter, cost



import logging
import os
import re
from pathlib import Path
from typing import Any

import numpy as np
from openai import AsyncOpenAI
from pydantic import BaseModel, Field

from nirvana.executors.constants import MODEL_PRICING, LLMProviders

logger = logging.getLogger(__name__)


def _get_api_key_from_file(file: Path) -> str:
    with open(file, 'r') as api_file:
        api_key = api_file.readline()
    return api_key


def _get_openai_compatible_provider_info(
    model_name: str,
    api_key: str | Path | None = None,
    base_url: str | None = None
) -> tuple[str, str]:
    """Infer provider-specific settings (API key and base URL) from the model name."""

    if isinstance(api_key, Path):
        api_key = _get_api_key_from_file(api_key)

    if model_name.startswith("gpt") or model_name.startswith("text-embedding"):
        base_url = LLMProviders.OPENAI
        api_key = api_key or os.getenv("OPENAI_API_KEY", "")
    elif model_name.startswith("deepseek"):
        base_url = LLMProviders.DEEPSEEK
        api_key = api_key or os.getenv("DEEPSEEK_API_KEY", "")
    elif model_name.startswith("qwen"):
        base_url = LLMProviders.QWEN
        api_key = api_key or os.getenv("QWEN_API_KEY", "")
    elif model_name.startswith("gemini"):
        base_url = LLMProviders.GEMINI
        api_key = api_key or os.getenv("GEMINI_API_KEY", "")
    else:
        raise ValueError(f"Unsupported model: {model_name}. Litellm will be used as the underlying backend in the next version.")

    return api_key, base_url


def _create_client(api_key: str, **kwargs):
    assert api_key != "", "API key is required."
    client = AsyncOpenAI(api_key=api_key, **kwargs)
    return client


class LLMArguments(BaseModel):
    max_tokens: int = Field(default=512, ge=1, le=16384, description="The maximum number of tokens to generate.")
    temperature: float = Field(default=0.1, ge=0.0, le=1.0, description="The sampling temperature.")
    max_timeouts: int = Field(default=3, ge=1, le=10, description="The maximum number of timeouts.")


class LLMClient:
    default_model: str | None = None
    client: AsyncOpenAI = None
    config: LLMArguments = LLMArguments()

    @classmethod
    def configure(
        cls,
        model_name: str,
        api_key: str | Path | None = None,
        base_url: str | None = None,
        **kwargs,
    ):
        """
        Configure the shared LLM client.

        The provider (OpenAI / DeepSeek / Qwen) is inferred from ``model_name``,
        and appropriate defaults for ``base_url`` and ``api_key`` are applied.
        Users can still override both ``api_key`` and ``base_url`` explicitly.
        """
        cls.default_model = model_name
        api_key, base_url = _get_openai_compatible_provider_info(
            model_name=model_name,
            api_key=api_key,
            base_url=base_url,
        )
        cls.client = _create_client(api_key=api_key, base_url=base_url, **kwargs)
        return cls()

    async def create_embedding(self, text: list[str] | str, embed_model: str = "text-embedding-3-large"):
        api_key, base_url = _get_openai_compatible_provider_info(model_name=embed_model)
        api_key = self.client.api_key if api_key == "" else api_key
        self.client = _create_client(api_key=api_key, base_url=base_url)
        response = await self.client.embeddings.create(
            input=text, model=embed_model
        )
        cost = (response.usage.total_tokens / 1000) * MODEL_PRICING[embed_model]["Input"]
        return np.array([data.embedding for data in response.data]).squeeze(), cost

    async def __call__(self,
        messages: str | list[dict[str, str]],
        parse_tags: bool = False,
        parse_code: bool = False,
        **kwargs,
    ) -> dict[str, Any]:
        model_name = kwargs.pop("model", None)
        if model_name is not None:
            api_key, base_url = _get_openai_compatible_provider_info(model_name=model_name)
            api_key = self.client.api_key if api_key == "" else api_key
            self.client = _create_client(api_key=api_key, base_url=base_url)
        else:
            model_name = self.default_model
        timeout = 0
        success = False
        while not success and timeout < self.config.max_timeouts:
            timeout += 1
            try:
                response = await self.client.responses.create(
                    model=model_name,
                    input=messages,
                    max_output_tokens=self.config.max_tokens,
                    temperature=self.config.temperature,
                )
                llm_output = response.output_text
                token_cost = self._compute_usage(response)
                success = True
            except Exception as e:
                logger.error(f"An error occurs when creating a response: {e}")

        outputs = dict()
        outputs["raw_output"] = llm_output
        if parse_tags:
            tags: list[str] = kwargs["tags"]
            for tag in tags:
                outputs[tag] = self._extract_xml(llm_output, tag)
        elif parse_code:
            code = self._extract_code(llm_output, lang=kwargs["lang"])
            outputs["output"] = code
        else:
            outputs["output"] = llm_output
        outputs["cost"] = token_cost
        return outputs
    
    def _compute_usage(self, response):
        """
        Compute the token cost of an LLM completion.
        
        Args:
            response: The response object from the LLM API.
        
        Returns:
            float: The total token cost.
        """
        model_name = response.model
        input_tokens = response.usage.input_tokens
        output_tokens = response.usage.output_tokens
        cached_tokens = response.usage.input_tokens_details.cached_tokens
        
        # Get the pricing for the model
        pricing = MODEL_PRICING[model_name]
        
        # Compute the cost for each type of token
        # If the model is qwen-series, no price differences are in input and cached tokens
        if model_name.startswith("qwen"):
            input_cost = (input_tokens / 1000) * pricing["Input"]
            output_cost = (output_tokens / 1000) * pricing["Output"]
            return input_cost + output_cost
        else:
            # Otherwise, consider the difference between input and cached tokens
            input_cost = (input_tokens - cached_tokens) / 1000 * pricing["Input"]
            output_cost = (output_tokens / 1000) * pricing["Output"]
            cache_cost = (cached_tokens / 1000) * pricing["Cache"]
            return input_cost + output_cost + cache_cost

    def _extract_xml(self, text: str, tag: str):
        match = re.search(f"<{tag}>(.*?)</{tag}>", text, re.DOTALL)
        return match.group(1).strip() if match else None
    
    def _extract_code(self, text: str, lang: str = "python"):
        pattern = rf"```{lang}(.*?)```"
        match = re.search(pattern, text, re.DOTALL)
        return match.group(1).strip() if match else None



from typing import Any, Iterable


class ReducePrompter:
    """
    TODO: Simple implementation that does not consider the case where the input length exceeds the token limit.
    The next step is to implement several optimizations, like `summarize and aggregate` and `incremental aggregation`
    """
    
    def __init__(self):
        self.system_instruction = (
            "You are a helpful assistant helping the user make sense of their data. "
            "You are performing a reduce operation (multiple inputs: one output) to "
            "aggregate the given data based on the user's instruction.\n"
            "Output the result of the reduce operation concisely in the following format.\n"
            "<output> LLM output </output>\n"
        )

    def generate_prompt(
            self, 
            data_set: Iterable[Any],
            user_instruction: str,
            dtype: str = "text",
    ):
        # 1. Prepare system message
        sys_message = [{"role": "system", "content": self.system_instruction}]

        # 2. Prepare user message
        user_content = []
        for ctr, data in enumerate(data_set):
            if dtype == "text" or dtype == "numeric":
                user_content.append({"type": "input_text", "text": f"Data {ctr + 1}: {str(data)}"})
            elif dtype == "image":
                user_content.append({"type": "input_text", "text": f"Data {ctr + 1}: "})
                user_content.append({"type": "input_image", "image_url": data})
            else:
                raise ValueError(f"Data type {dtype} is not supported.")
        user_content.append({"type": "input_text", "text": user_instruction})
        user_message = [{"role": "user", "content": user_content}]

        messages = sys_message + user_message
        return messages
